import json
import os
import gradio as gr
import numpy as np
import umap
import matplotlib.pyplot as plt
from io import BytesIO
from PIL import Image


def init_umap_input_list(storage_path, project_name):
    if project_name == None:
        return gr.update(choices=[])

    project_path = os.path.join(storage_path, project_name)
    project_config_path = os.path.join(project_path, 'config.json')
    project_config = json.load(open(project_config_path, 'r'))
    umap_input_list = [k for k, v in project_config['umap_input'].items()]

    return gr.update(choices=umap_input_list)


def umap_dimension_reduction(X, layer_1_neighber, layer_2_neighber):
    
    layer_1_neighber, layer_2_neighber = int(layer_1_neighber), int(layer_2_neighber)
    print('layer_1_neighber, layer_2_neighber',layer_1_neighber, layer_2_neighber)
    if layer_2_neighber == 0:
        reducer = umap.UMAP(n_neighbors=layer_1_neighber, min_dist=0.1, n_components=2)
        data_2d = reducer.fit_transform(X)
        return data_2d, [reducer]

    if len(X) < layer_1_neighber or len(X) < layer_2_neighber:
        gr.Info("neighber should smaller than len(X)")
        return [], []
    reducer_1 = umap.UMAP(n_neighbors=layer_1_neighber, min_dist=0.3, n_components=100)
    reducer_2 = umap.UMAP(n_neighbors=layer_2_neighber, min_dist=0.1, n_components=2)
    print('X.shape', X.shape)
    # print(X[0])
    Z = reducer_1.fit_transform(X)
    Z = reducer_2.fit_transform(Z)
    return Z, [reducer_1, reducer_2]


def plot_on_img(X_2d):
    if len(X_2d) == 0:
        X_2d = np.array([[0, 0]])

    # 在散點圖中添加 X 和 Y 軸的比例尺
    fig, ax = plt.subplots(figsize=(7.2, 7.2), dpi=100)
    plt.subplots_adjust(left=0, right=1, top=1, bottom=0)

    # 繪製散點圖
    x = X_2d[:, 0]
    y = X_2d[:, 1]
    c = np.arange(len(X_2d))
    ax.scatter(x, y, c=c)

    # 隱藏坐標軸標籤
    ax.axis('off')
    # 將圖像轉換為 PIL.Image 對象
    buf = BytesIO()
    fig.savefig(buf, format='png', dpi=100, bbox_inches='tight', pad_inches=0)
    buf.seek(0)
    plt.close()
    return Image.open(buf)



def generate_umap_gallery(storage_path, project_name, umap_input):
    project_path = os.path.join(storage_path, project_name)
    project_config_path = os.path.join(project_path, 'config.json')
    project_config = json.load(open(project_config_path, 'r'))
    latent_dir_path = os.path.join(storage_path, project_name, 'latent')
    latent_list = project_config['umap_input'][umap_input]['latent']
    umap_input_dir_path = os.path.join(project_path, 'umap_input')
    umap_input_path = os.path.join(umap_input_dir_path, umap_input)
    umap_input_mask = np.load(umap_input_path)['umap_mask']
    print("umap_input_mask",umap_input_mask.shape)
    # print(umap_input_mask)
    multi_latent = np.zeros((0, project_config['observer_dim']))
    print('latent_list', len(latent_list))
    for it in latent_list:
        latent_path = os.path.join(latent_dir_path, it)
        latent = np.load(latent_path)['latent']
        multi_latent = np.append(multi_latent, latent, axis=0)


    print('multi_latent',multi_latent.shape)
    umap_list = []
    img_list = []
    for i in [15, 30, 50, 200, 500]:
        for j in [0, 10, 20]:
            X_2d = np.zeros((len(multi_latent), 2)) + np.nan
            try:
                X_2d[umap_input_mask], umap = umap_dimension_reduction(multi_latent[umap_input_mask], i, j)
                img_list.append(plot_on_img(X_2d))
                umap_list.append(umap)
            except:
                pass
    return img_list, umap_list



def create_dimension_reduction_ui(storage_path, project_name, init_tab):
    ui = dict()

    umap_list = gr.State(None)
    with gr.Row(visible=True):
        with gr.Column(scale=2):
            ui['input_list'] = gr.Dropdown(label='Input List')
            ui['umap_library'] = gr.Dropdown(label='UMAP Library', choices=['umap_learn (on CPU)'], value="umap_learn (on CPU)")
            ui['umap_windown_size'] = gr.Number(label='UMAP Window size', precision=0)
            ui['generate_umap_gallery'] = gr.Button('Generate Some UMAP Result')
        with gr.Column(scale=8):
            ui['umap_gallery'] = gr.Gallery(label='UMAP Gallery', columns=5)
    



    init_tab.select(
        fn=init_umap_input_list,
        inputs=[storage_path, project_name],
        outputs=ui['input_list']
    )

    ui['generate_umap_gallery'].click(
        fn=generate_umap_gallery,
        inputs=[storage_path, project_name, ui['input_list']],
        outputs=[ui['umap_gallery'], umap_list]
    )

    return ui


import gradio as gr
import os
import json
import numpy as np

from castle import generate_dinov2
from castle.utils.video_io import ReadArray
import torch
from torch.utils.data import Dataset, DataLoader

import torchvision.transforms as tt

from castle.utils.h5_io import H5IO
from castle.utils.video_align import center_roi, rotate_based_on_roi_closest_center_point, crop, blank_page
from castle.utils.plot import generate_mix_image

def init_select_video_list(storage_path, project_name):
    if project_name == None:
        return gr.update(choices=[])
    project_path = os.path.join(storage_path, project_name)
    project_config_path = os.path.join(project_path, 'config.json')
    project_config = json.load(open(project_config_path, 'r'))
    if not 'source' in project_config:
         return gr.update(choices=[])
    
    subdirectories = sorted(project_config['source'])
    subdirectories.append("All")
    return gr.update(choices=subdirectories)


def extract_roi_latent(storage_path, project_name, select_model, select_roi, select_video, batch_size, preprocess, progress=gr.Progress()):
    batch_size = int(batch_size)
    project_path = os.path.join(storage_path, project_name)
    project_config_path = os.path.join(project_path, 'config.json')
    project_config = json.load(open(project_config_path, 'r'))
    latent_dir_path = os.path.join(storage_path, project_name, 'latent')
    os.makedirs(latent_dir_path, exist_ok=True)
    observer = generate_dinov2(model_type='dinov2_vitb14_reg')
    project_config['observer_dim'] = observer.n_feature
    latent_file_list = []
    select_roi = int(select_roi)
    
    if select_video == "All":
        subdirectories = sorted(project_config['source'])

    else:
        subdirectories = [select_video]

    for it in subdirectories:
        source_video = ReadArray(os.path.join(storage_path, project_name, 'sources', it))
        track_dir_path = os.path.join(project_path, 'track', it)
        mask_list_path = os.path.join(track_dir_path, f'mask_list.h5')
        tracker = H5IO(mask_list_path)
        latent = extract_roi_latent_from_video(observer, source_video, tracker, batch_size, select_roi, preprocess, progress)
        # try:
        #     latent = extract_roi_latent_from_video(observer, source_video, tracker, batch_size, select_roi, progress)
        # except:
        #     gr.Info("latent extract fail")
        #     del observer
        #     return []
        base_name = source_video.video_name.split('.')[0]
        latent_path = os.path.join(latent_dir_path, f'{base_name}_ROI_{select_roi}_latent.npz')
        # print('latent', type(latent))
        np.savez_compressed(latent_path, latent=latent)

        if not 'latent' in project_config:
            project_config['latent'] = dict()
        
        project_config['latent'][f'{base_name}_ROI_{select_roi}_latent.npz'] = source_video.video_name
        json.dump(project_config, open(project_config_path,'w'))
        latent_file_list.append(latent_path)
    del observer
    del tracker
    return latent_file_list


# class CustomImageDataset(Dataset):
#     def __init__(self, source_video, tracker):
#         self.source_video = source_video
#         self.tracker = tracker

#     def __len__(self):
#         return min(self.tracker.read_config('total_frames'), len(self.source_video))

#     def __getitem__(self, index):
#         return self.source_video[index], self.tracker.read_mask(index)
    


def extract_roi_latent_from_video(observer, source_video, tracker, batch_size, select_roi, preprocess, progress):

    latent_list = []
    batch_size = int(batch_size)
    print('batch_size', (batch_size))


    for i in progress.tqdm(range(0, len(source_video), batch_size)):
        frames = [source_video[i+j] for j in range(batch_size) if j+i < len(source_video)]
        masks = [tracker.read_mask(i+j) for j in range(batch_size) if j+i < len(source_video)]
        
        frames_process, masks_process = [], []
        for f, m in zip(frames, masks):
            ff, mm = preprocess.transform(f, m)
            frames_process.append(ff)
            masks_process.append(mm)

        frames = frames_process
        masks = masks_process

        try:
            latent = observer.extract_batch_latent(frames, masks, select_roi)
            latent_list.extend(latent)
            continue
        except:
            print(f"batch {i} error, try to run frame one by one")

        for j in range(batch_size):
            try:
                latent = observer.extract_image_latent(frames[j], masks[j], select_roi)
                latent_list.append(latent)
            except:
                latent_list.append(observer.nan_latent())
                print(f'fail at frame {i*batch_size+j}')

    latent_list = np.array(latent_list)
    print('latent_list', latent_list.shape)
    return latent_list


class Preprocess:
    def __init__(self, center_roi_switch, center_roi_id, center_roi_crop_width, center_roi_crop_height,  rotate_roi_tail_switch, rotate_roi_tail_id):
        self.center_roi_switch = center_roi_switch
        self.center_roi_id=center_roi_id
        self.center_roi_crop_width=center_roi_crop_width
        self.center_roi_crop_height=center_roi_crop_height
        self.rotate_roi_tail_switch=rotate_roi_tail_switch
        self.rotate_roi_tail_id=rotate_roi_tail_id


    def transform(self, frame, mask):

        try:
            if self.center_roi_switch == 'True':
                f = center_roi(frame, mask, self.center_roi_id)
                m = center_roi(mask, mask, self.center_roi_id)
                if self.rotate_roi_tail_switch == 'True':
                    f = rotate_based_on_roi_closest_center_point(f, m, self.rotate_roi_tail_id)
                    m = rotate_based_on_roi_closest_center_point(m, m, self.rotate_roi_tail_id)
                f = crop(f, self.center_roi_crop_height, self.center_roi_crop_width)
                m = crop(m, self.center_roi_crop_height, self.center_roi_crop_width)
            else:
                f, m = frame, mask
        except:
            f = blank_page(self.center_roi_crop_height, self.center_roi_crop_width)
            m = blank_page(self.center_roi_crop_height, self.center_roi_crop_width)

        return f, m

def setting_preprocess(storage_path, project_name, select_video, center_roi_switch, center_roi_id, center_roi_crop_width, center_roi_crop_height,  rotate_roi_tail_switch, rotate_roi_tail_id):
    preprocess = Preprocess(center_roi_switch, center_roi_id, center_roi_crop_width, center_roi_crop_height,  rotate_roi_tail_switch, rotate_roi_tail_id)

    project_path = os.path.join(storage_path, project_name)
    project_config_path = os.path.join(project_path, 'config.json')
    project_config = json.load(open(project_config_path, 'r'))
    latent_dir_path = os.path.join(storage_path, project_name, 'latent')
    os.makedirs(latent_dir_path, exist_ok=True)
    observer = generate_dinov2(model_type='dinov2_vitb14_reg')
    project_config['observer_dim'] = observer.n_feature

    if select_video == "All":
        subdirectories = sorted(project_config['source'])
    else:
        subdirectories = [select_video]
    print('subdirectories', subdirectories)
    first_frame = ReadArray(os.path.join(storage_path, project_name, 'sources', subdirectories[0]))[0]
    track_dir_path = os.path.join(project_path, 'track', subdirectories[0])
    mask_list_path = os.path.join(track_dir_path, f'mask_list.h5')
    tracker = H5IO(mask_list_path)
    f, m = preprocess.transform(first_frame, tracker.read_mask(0))

    return preprocess, generate_mix_image(f, m)


def create_extract_ui(storage_path, project_name, extract_tab):
    ui = dict()
    preprocess = gr.State(None)
    with gr.Row(visible=True):
        with gr.Column(scale=2):
            ui['select_model'] = gr.Dropdown(
                label="Select Visual Model", 
                choices=["dinov2_vitb14_reg4_pretrain"], value="dinov2_vitb14_reg4_pretrain", visible=False)
            ui['select_roi_id'] = gr.Textbox(label="Enter ROI ID", value="1", info="ex: 1,2,3.", visible=False)
            ui['batch_size'] = gr.Textbox(label="Batch size", value="16", info="ex: 1, 16, 64 ...", visible=False)
            ui['select_video'] = gr.Dropdown(label="Select Target Video", value='All', visible=False)
            
            # ui['progress_bar'] = gr.Textbox(label="Progress", visible=False)
        with gr.Column(scale=2):
            ui['center_roi_switch'] =gr.Dropdown(label="Center ROI", value='False', choices=['True', 'False'], visible=False)
            ui['center_roi_id'] =gr.Number(label="Center ROI ID", value=1, visible=False)
            ui['center_roi_crop_width'] =gr.Number(label="width", value=300, visible=False)
            ui['center_roi_crop_height'] =gr.Number(label="height", value=300, visible=False)
            ui['rotate_roi_tail_switch'] =gr.Dropdown(label="Rotate based on Tail", value='False', choices=['True', 'False'], visible=False)
            ui['rotate_roi_tail_id'] =gr.Number(label="Tail ROI ID", value=2, visible=False)
            ui['apply_preprocess'] = gr.Button("Apply", visible=False)
            
        with gr.Column(scale=4):
            ui['display'] = gr.Image(
                    label='Display', interactive=False, visible=False)
            ui['extract_btn'] = gr.Button("Extract", visible=False)
            ui['latent_file_list'] = gr.File(label="ROI Visual Representation File List", visible=False)
            

    extract_tab.select(
        fn=init_select_video_list,
        inputs=[storage_path, project_name],
        outputs=ui['select_video']
    )

    ui['extract_btn'].click(
        fn=extract_roi_latent,
        inputs=[storage_path, project_name, ui['select_model'], ui['select_roi_id'], ui['select_video'], ui['batch_size'], preprocess],
        outputs=ui['latent_file_list']
    )
    ui['apply_preprocess'].click(
        fn=setting_preprocess,
        inputs=[storage_path, project_name, ui['select_video'], ui['center_roi_switch'], ui['center_roi_id'],ui['center_roi_crop_width'],ui['center_roi_crop_height'],  ui['rotate_roi_tail_switch'],ui['rotate_roi_tail_id'] ],
        outputs=[preprocess, ui['display']]
    )
    return ui



import gradio as gr
import os
import cv2
import h5py
import numpy as np
import av

from .plot_mask_info import Plotter
from castle.utils.plot import generate_mix_image, generate_mask_image
# from api.segmentor import merge_frame_and_mask, mask2img
# from media.tracking_io import TrackingIO
from castle.utils.h5_io import H5IO

def plot_basic_mask_info(storage_path, project_name, source_video, progress=gr.Progress()):
    project_path = os.path.join(storage_path, project_name)
    video_name = source_video.video_name
    track_dir_path = os.path.join(project_path, 'track', video_name)
    mask_list_path = os.path.join(track_dir_path, f'mask_list.h5')
    tracker = H5IO(mask_list_path)

    with h5py.File(mask_list_path, 'r') as f:
        roi_count = tracker.read_config('roi_count')
        total_frames = tracker.read_config('total_frames')
        roi_info_list = [{"x":[], "y":[], "area":[]} for i in range(roi_count)]
        for i_frame in progress.tqdm(range(total_frames)):
            for i in range(roi_count):
                if f"{i_frame}" in f:
                    mask = f[str(i_frame)][:]
                    # print('mask', mask.shape)
                    mask = cv2.inRange(mask, i+1, i+1)
                    output = cv2.connectedComponentsWithStats(mask, 8, cv2.CV_32S)
                    num_labels, _, stats, centroids = output
                    if num_labels > 1:
                        areas = [stats[j, cv2.CC_STAT_AREA] for j in range(1, num_labels)]
                        max_label = np.argmax(areas)
                        roi_info_list[i]['x'].append(round(centroids[max_label + 1][0]))
                        roi_info_list[i]['y'].append(round(centroids[max_label + 1][1]))
                        roi_info_list[i]['area'].append(areas[max_label])
                        continue
                roi_info_list[i]['x'].append(np.nan)
                roi_info_list[i]['y'].append(np.nan)
                roi_info_list[i]['area'].append(0)


    # print(roi_info_list)
    for i in range(roi_count):
        roi_info_list[i]['x'] = np.array(roi_info_list[i]['x'])
        roi_info_list[i]['y'] = np.array(roi_info_list[i]['y'])
        roi_info_list[i]['area'] = np.array(roi_info_list[i]['area']).astype(int)

    del tracker
    return Plotter.plot_position(roi_info_list), Plotter.plot_speed(roi_info_list), Plotter.plot_area(roi_info_list), roi_info_list


def generate_mask_kinematic_csv(storage_path, project_name, source_video, roi_info_list):
    project_path = os.path.join(storage_path, project_name)
    video_name = source_video.video_name
    track_dir_path = os.path.join(project_path, 'track', video_name)
    video_name_wo_extension = video_name.split('.')[0]
    mask_kinematic_csv_path = os.path.join(track_dir_path, f'{video_name_wo_extension}-basic-infomation.csv')
    df = Plotter.create_pandas(roi_info_list)
    df.to_csv(mask_kinematic_csv_path)
    return mask_kinematic_csv_path


def generate_mask_video(storage_path, project_name, source_video):
    project_path = os.path.join(storage_path, project_name)
    video_name = source_video.video_name
    track_dir_path = os.path.join(project_path, 'track', video_name)
    mask_list_path = os.path.join(track_dir_path, f'mask_list.h5')
    video_name_wo_extension = video_name.split('.')[0]
    mask_video_path = os.path.join(track_dir_path, f'{video_name_wo_extension}-mask.mp4')
    output = av.open(mask_video_path, 'w')

    # Add a stream, specify codec and settings
    stream = output.add_stream('libx264', rate=source_video.fps)
    stream.options = {'crf': '22'}
    stream.pix_fmt = 'yuv420p'  # Most common pixel format
    
    tracker = H5IO(mask_list_path)
    stream.height = tracker.read_config('height')
    stream.width = tracker.read_config('width')
    total_frames = tracker.read_config('total_frames')

    for i_frame in range(total_frames):
        it = tracker.read_mask(i_frame)
        frame = av.VideoFrame.from_ndarray(generate_mask_image(it), format='rgb24')
        for packet in stream.encode(frame):
            output.mux(packet)

    for packet in stream.encode():
        output.mux(packet)

    output.close()
    del tracker
    return mask_video_path

def generate_mix_video(storage_path, project_name, source_video):
    project_path = os.path.join(storage_path, project_name)
    video_name = source_video.video_name
    track_dir_path = os.path.join(project_path, 'track', video_name)
    mask_list_path = os.path.join(track_dir_path, f'mask_list.h5')
    tracker = H5IO(mask_list_path)

    video_name_wo_extension = video_name.split('.')[0]
    mix_video_path = os.path.join(track_dir_path, f'{video_name_wo_extension}-mix.mp4')
    output = av.open(mix_video_path, 'w')

    # Add a stream, specify codec and settings
    stream = output.add_stream('libx264', rate=source_video.fps)
    stream.options = {'crf': '22'}

    stream.height = tracker.read_config('height')
    stream.width = tracker.read_config('width')
    total_frames = tracker.read_config('total_frames')
    stream.pix_fmt = 'yuv420p'  # Most common pixel format
    for i in range(total_frames):
        it = tracker.read_mask(i)
        frame = source_video[i]
        mix = generate_mix_image(frame, it)
        mix = av.VideoFrame.from_ndarray(mix, format='rgb24')
        for packet in stream.encode(mix):
            output.mux(packet)

    for packet in stream.encode():
        output.mux(packet)

    output.close()
    del tracker
    return mix_video_path

def create_post_track_ui(storage_path, project_name, source_video):
    ui = dict()
    roi_info_list = gr.State(None)
    # with gr.Accordion('Basic Kinematic Infomation', open=True, visible=False) as ui['basic_mask_info_accordion']:
    ui['analysis_mask'] = gr.Button("Analysis Mask", interactive=True, visible=False)
    with gr.Row(visible=True):
        with gr.Tab(label='position'):
            ui['position_plot'] = gr.Plot(label="Position", visible=False)
        with gr.Tab(label='speed'):
            ui['velocity_plot'] = gr.Plot(label="Speed", visible=False)
        with gr.Tab(label='area'):
            ui['area_plot'] = gr.Plot(label="Area", visible=False)  
    with gr.Row(visible=True):
        with gr.Column(scale=2):
            ui['generate_mask_kinematic_btn'] = gr.Button("Generate Basic Kinematic CSV", interactive=True, visible=False)
        with gr.Column(scale=8):
            ui['mask_kinematic_file'] = gr.File(label="Basic Kinematic CSV", interactive=False, visible=False)
    with gr.Row(visible=True):
        with gr.Column(scale=2):
            ui['generate_mask_video_btn'] = gr.Button("Generate Mask Video", interactive=True, visible=False)
        with gr.Column(scale=8):
            ui['mask_video'] = gr.File(label="Mask Video", interactive=False, visible=False)
    with gr.Row(visible=True):
        with gr.Column(scale=2):
            ui['generate_mix_video_btn'] = gr.Button("Generate Mix Video", interactive=True, visible=False)
        with gr.Column(scale=8):
            ui['mix_video'] = gr.File(label="Mix Video", interactive=False, visible=False)

    ui['analysis_mask'].click(
        fn=plot_basic_mask_info,
        inputs=[storage_path, project_name, source_video],
        outputs=[ui['position_plot'], ui['velocity_plot'], ui['area_plot'], roi_info_list]
    )

    ui['generate_mask_kinematic_btn'].click(
        fn=generate_mask_kinematic_csv,
        inputs=[storage_path, project_name, source_video, roi_info_list],
        outputs=ui['mask_kinematic_file']
    )

    ui['generate_mask_video_btn'].click(
        fn=generate_mask_video,
        inputs=[storage_path, project_name, source_video],
        outputs=ui['mask_video']
    )

    ui['generate_mix_video_btn'].click(
        fn=generate_mix_video,
        inputs=[storage_path, project_name, source_video],
        outputs=ui['mix_video']
    )
      
    return ui


import torch


def generate_permute_matrix(dim, num, keep_first=True, gpu_id=0):
    all_matrix = []
    for idx in range(num):
        random_matrix = torch.eye(dim, device=torch.device('cuda', gpu_id))
        if keep_first:
            fg = random_matrix[1:][torch.randperm(dim - 1)]
            random_matrix = torch.cat([random_matrix[0:1], fg], dim=0)
        else:
            random_matrix = random_matrix[torch.randperm(dim)]
        all_matrix.append(random_matrix)
    return torch.stack(all_matrix, dim=0)


def truncated_normal_(tensor, mean=0, std=.02):
    size = tensor.shape
    tmp = tensor.new_empty(size + (4, )).normal_()
    valid = (tmp < 2) & (tmp > -2)
    ind = valid.max(-1, keepdim=True)[1]
    tensor.data.copy_(tmp.gather(-1, ind).squeeze(-1))
    tensor.data.mul_(std).add_(mean)
    return tensor



import gradio as gr
import os
import numpy as np

from castle.utils.plot import generate_mask_image, generate_mix_image
from castle.utils.h5_io import H5IO

def index_slide_apply(storage_path, project_name, source_video, index, mode):
    if mode == 'Image':
        return source_video[index]
    
    project_path = os.path.join(storage_path, project_name)
    video_name = source_video.video_name
    track_dir_path = os.path.join(project_path, 'track', video_name)
    mask_list_path = os.path.join(track_dir_path, f'mask_list.h5')

    tracker = H5IO(mask_list_path)

    if mode == 'Image & Mask':
        frame = source_video[index]
        mix = generate_mix_image(frame, tracker.read_mask(index))
        return mix
    if mode == 'Mask':
        return generate_mask_image(tracker.read_mask(index))
    
    del tracker



def create_view_ui(storage_path, project_name, source_video):
    ui = dict()
    with gr.Row(visible=True):
        with gr.Column(scale=2):
            ui['display_mode_drop'] = gr.Dropdown(
                    choices=["Image", "Image & Mask", "Mask"],
                    label="Display Mode", value="Image",
                    interactive=True, visible=False)

            ui['step_frame'] = gr.Dropdown(
                    choices=["1", "5", "10", "100", "1000", "10000"],
                    label="Frame step number", value="1",
                    interactive=True, visible=False)
            
        with gr.Column(scale=8):
            ui['display_view'] = gr.Image(label='Display', interactive=False, visible=False)
            ui['index_slide'] = gr.Slider(label="Frame", minimum=0, step=1, maximum=1, value=0, interactive=True, visible=False)
          
    ui['index_slide'].change(
        fn=index_slide_apply,
        inputs=[storage_path, project_name, source_video, ui['index_slide'], ui['display_mode_drop']],
        outputs=ui['display_view']
    )
    step_frame_apply = lambda step_frame: gr.update(step=step_frame)
    ui['step_frame'].change(
        fn=step_frame_apply,
        inputs=ui['step_frame'],
        outputs=ui['index_slide']
    )
    return ui
    


import os

import json
import h5py
import time
import glob
import gradio as gr
import numpy as np
from natsort import natsorted
from castle import generate_aot
from castle.utils.plot import generate_mix_image
from castle.utils.h5_io import H5IO

# from api.segmentor import merge_frame_and_mask

# from api.tracker import get_aot
# from media.tracking_io import TrackingIO

# model_config = json.load(open('config/model_config.json', 'r'))


def read_label(storage_path, project_name, source_video):
    if source_video == None:
        return []
    project_path = os.path.join(storage_path, project_name)
    video_name = source_video.video_name
    label_dir_path = os.path.join(project_path, 'label', video_name)
    label_file_list = natsorted(glob.glob(f'{label_dir_path}/*.npz'))
    print('label_file_list', label_file_list)
    label_list = []
    for it in label_file_list:
        index = os.path.basename(it).split('.')[0]
        f = np.load(it)
        frame, mask = f['frame'], f['mask']
        label = dict()
        label['index'] = index
        label['frame'] = frame
        label['mask'] = mask

        label_list.append(label)

    return label_list


def read_label_to_gallery(storage_path, project_name, source_video):
    label_list = read_label(storage_path, project_name, source_video)
    gallery_list = []
    for it in label_list:
        index = it['index']
        frame, mask = it['frame'], it['mask']
        mix = generate_mix_image(frame, mask)
        gallery_list.append((mix, index))

    return label_list, gallery_list


# def setting_start_frame(label_list, evt: gr.SelectData):
#     label_index = evt.index
#     frame_index = label_list[label_index]['index']
#     return frame_index


# def collapse_accordion():
#     return gr.update(open=False)


# def open_accordion():
#     return gr.update(open=True)

# def init_display(label_list, evt: gr.SelectData):
#     label_index = evt.index
#     frame, mask = label_list[label_index]['frame'], label_list[label_index]['mask']
#     mix = generate_mix_image(frame, mask)
#     return mix



class Interfence:
    def __init__(self, storage_path, project_name, source_video, start, stop, max_len, model_aot):
        self.cancel = False
        self.show_middle_result = False
        self.model_aot = model_aot

        project_path = os.path.join(storage_path, project_name)
        video_name = source_video.video_name
        track_dir_path = os.path.join(project_path, 'track', video_name)
        os.makedirs(track_dir_path, exist_ok=True)
        self.track_dir_path = track_dir_path

        label_path = os.path.join(project_path, 'label', video_name, f'{start}.npz')
        # self.first_label = np.load(label_path)
        self.source_video = source_video
        self.start = int(start)
        self.stop = int(stop)
        self.max_len = int(max_len)

        self.knowledges = []
        label_list = read_label(storage_path, project_name, source_video)
        self.roi_count = 0
        for it in label_list:
            index = it['index']
            frame, mask = it['frame'], it['mask']
            self.knowledges.append((frame, mask))
            self.roi_count = max(self.roi_count, np.max(mask))

        
        pass



    def tracking(self, progress):
        
        time.sleep(1)
        start, stop, max_len = self.start, self.stop, self.max_len,
        tracker = generate_aot(model_type=self.model_aot)
        # self.roi_count = np.max(self.first_label['mask'])
        self.write_mask_config()
        for f, m in self.knowledges:
            tracker.add_reference_frame(f, m, self.roi_count, -1)


        # tracker.add_reference_frame(self.first_label['frame'], self.first_label['mask'], self.roi_count, -1)
        delta = 1 if start < stop else -1
  

        for i in progress.tqdm(range(start, stop + delta, delta)):
            if self.cancel:
                self.show_middle_result = False
                self.cancel = False
                return "Cancel"
            
            frame = self.source_video[i]
            mask = tracker.track(frame)
            tracker.update_memory(mask)
            self.frame, self.mask = frame, mask.squeeze().detach().cpu().numpy().astype(np.uint8)
            self.write_mask(i, self.mask)

        self.show_middle_result = False
        return "Done"
    
    def write_mask_config(self):
        mask_list_path = os.path.join(self.track_dir_path, f'mask_list.h5')       


        tracker = H5IO(mask_list_path)
        tracker.write_config('roi_count', self.roi_count)
        tracker.write_config('total_frames', len(self.source_video))
        tracker.write_config('height', self.source_video.video_stream.height)
        tracker.write_config('width', self.source_video.video_stream.width)
        del tracker


    def write_mask(self, index, mask):
        mask_list_path = os.path.join(self.track_dir_path, f'mask_list.h5')            
        mode = 'a' if os.path.isfile(mask_list_path) else 'w'
        with h5py.File(mask_list_path, mode) as f:
            if f"{index}" in f:
                dset = f[f"{index}"][:]
                dset[:] = mask
            else:
                dset = f.create_dataset(f"{index}", mask.shape, dtype='uint8', compression="gzip", compression_opts=9)
                dset[:] = mask


    def set_cancel(self):
        self.cancel = True
        pass

    def flip_show_middle_result(self):
        if self.show_middle_result:
            self.show_middle_result = False
        else:
            self.show_middle_result = True


def init_Interfence(storage_path, project_name, source_video, start, stop, max_len, model_aot):
    print('init_Interfence', start, stop)
    return Interfence(storage_path, project_name, source_video, start, stop, max_len, model_aot)

def run_interfence(interfence, progress=gr.Progress()):
    status = interfence.tracking(progress)
    return f"{status}. From {interfence.start} to {interfence.stop}"

def click_middle_result(interfence):
    print('enter click_middle_result', interfence.show_middle_result)
    interfence.flip_show_middle_result()
    while interfence.show_middle_result:
        time.sleep(1)
        yield generate_mix_image(interfence.frame, interfence.mask), display_middle_result_mode(interfence.show_middle_result)

def display_middle_result_mode(res):
    if res:
       return "Show"
    else:
       return "Close"

def set_cancel(interfence):
    interfence.set_cancel()



def create_track_ui(storage_path, project_name, source_video, track_tab):
    ui = dict()

    label_list = gr.State(None)
    interfence = gr.State(None)
    with gr.Accordion('ROIs Knowledge', visible=False) as ui['gallery_accordion']:
        ui['gallery'] = gr.Gallery(
            label="Label Frame", show_label=True, allow_preview=False, object_fit="contain", columns=3)

    with gr.Accordion('Inference', open=True, visible=False) as ui['inference_accordion']:
        with gr.Row(visible=True):
            with gr.Column(scale=2):
                # ui['start_frame'] = gr.Textbox(
                #     label="Start Frame", interactive=False, visible=False)
                ui['start_frame'] = gr.Slider(
                    label="Start Frame (include)", minimum=0, step=1, maximum=1, value=0, interactive=True, visible=False)
                ui['stop_frame'] = gr.Slider(
                    label="Stop Frame (include)", minimum=0, step=1, maximum=1, value=1, interactive=True, visible=False)
                ui['deaot_model'] = gr.Dropdown(['r50_deaotl', 'swinb_deaotl'], label='tracking_model',
                              info='swim_transformer or res50',
                              value='r50_deaotl',
                              interactive=True)
                ui['long_term_max_len'] = gr.Number(
                    label="Long term menary length",
                    info="Bigger is better, but this depends on the GPU's RAM capacity",
                    value=30, interactive=True, visible=False
                )
                ui['init_tracker'] = gr.Button(
                    "init Tracker", interactive=True, visible=False)
                ui['tracking_btm'] = gr.Button(
                    "Tracking ROIs", interactive=True, visible=False)
                ui['progress_edit'] = gr.Textbox(
                    label="Progress", visible=False)
                ui['display_middle_result_mode'] = gr.Textbox(value="Close",
                    label="Display Mode", interactive=False, visible=False)
                ui['display_middle_result'] = gr.Button(
                    "Display middle result", interactive=True, visible=False)
                ui['cancel_btn'] = gr.Button(
                    "Cancel", interactive=True, visible=False)

            with gr.Column(scale=8):
                ui['display'] = gr.Image(
                    label='Display', interactive=False, visible=False)


    tracking_config = [ui['start_frame'], ui['stop_frame'], ui['long_term_max_len'], ui['deaot_model']]
    track_tab.select(
        fn=read_label_to_gallery,
        inputs=[storage_path, project_name, source_video],
        outputs=[label_list, ui['gallery']]
    )

    # ui['gallery'].select(
    #     fn=setting_start_frame,
    #     inputs=[label_list],
    #     outputs=ui['start_frame']
    # )
    # ui['gallery'].select(
    #     fn=collapse_accordion,
    #     outputs=ui['gallery_accordion']
    # )
    # ui['gallery'].select(
    #     fn=open_accordion,
    #     outputs=ui['inference_accordion']
    # )

    # ui['gallery'].select(
    #     fn=init_display,
    #     inputs=label_list,
    #     outputs=ui['display']
    # )

    init_Interfence_input = [storage_path, project_name, source_video]
    init_Interfence_input.extend(tracking_config)
    ui['init_tracker'].click(
        fn=init_Interfence,
        inputs=init_Interfence_input,
        outputs= interfence
    )

    ui['tracking_btm'].click(
        fn=run_interfence,
        inputs = interfence,
        outputs=ui['progress_edit']
    )

    ui['display_middle_result'].click(
        fn=click_middle_result,
        inputs=interfence,
        outputs=[ui['display'], ui['display_middle_result_mode']]
    )

    ui['cancel_btn'].click(
        fn=set_cancel,
        inputs=interfence,
    )

    return ui



import os
import json
import shutil
import datetime
import gradio as gr

def lock_project_page(object_count):
    return [gr.update(interactive=False) for i in range(object_count)]

def unlock_old_project_btn(object_count):
    return [gr.update(interactive=True) for i in range(object_count)]

def list_project(storage_path):
    li = [d for d in os.listdir(storage_path) if os.path.isdir(os.path.join(storage_path, d))]
    if len(li) > 1:
        subdirectories = sorted(li)
    else:
        subdirectories = li
    return gr.update(choices=subdirectories)

def create_new_project(storage_path, project_name):
    project_path = os.path.join(storage_path, project_name)
    project_config_path = os.path.join(project_path, 'config.json')
    os.makedirs(project_path)
    project_config = dict()
    project_config['project_name'] = project_name
    json.dump(project_config, open(project_config_path,'w'))

def default_new_project_name(new_project_name):
    if len(new_project_name) > 0:
        return gr.update(value=new_project_name)
    
    create_time = datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
    return gr.update(value=f'{create_time}-Project')

def set_project_name(project_name):
    return project_name

def delete_old_project(storage_path, project_name):
    folder_path = os.path.join(storage_path, project_name)
    if os.path.exists(folder_path):
        print('folder_path', folder_path)
        shutil.rmtree(folder_path)
        gr.Info(f"Deleted {project_name}")
        print("Deleted")
    else:
        print("No such folder")



def create_project_ui(OS_SYS, root=''):
    if root == None:
        root = os.path.join('projects','')
    sep = os.sep
    if not root.endswith(sep):
        root += sep
    if not os.path.exists(root):
        os.makedirs(root)
        
    ui = dict()
    DEFAULT_DEVICE = 'mps' if OS_SYS == 'Darwin' else 'cuda'
    ui['device'] = gr.Textbox(label='Device',
                              info='For example: mps, cpu, cuda, or cuda:x',
                              value=DEFAULT_DEVICE,
                              interactive=False, visible=False)

    ui['storage_path'] = gr.Textbox(label='Storage Location',
                        info='The location which storage all project',
                        value=root,
                        interactive=True)
    
    with gr.Tab(label='Open Old Project'):
        ui['old_project_drop'] = gr.Dropdown(label="Open Project", interactive=True)
        with gr.Row(visible=True):
            with gr.Column(scale=9):
                ui['old_project_open_btn'] = gr.Button('Open', interactive=False)
            with gr.Column(scale=1):
                ui['old_project_delete_btn'] = gr.Button('Delete', interactive=False)
    with gr.Tab(label='New Project'):
        ui['new_project_name'] = gr.Textbox(label='New Project Name', interactive=True)
        ui['new_project_create_btn'] = gr.Button('Create', interactive=True)

    
    ui['project_name'] = gr.State(None)
    
    object_count = gr.State(len(ui))
    old_project_btn_list = [ui['old_project_open_btn'], ui['old_project_delete_btn']]
    old_project_btn_count = gr.State(len(old_project_btn_list))


    ui['new_project_create_btn'].click(
        fn=lock_project_page,
        inputs=object_count,
        outputs=[v for k, v in ui.items()]
    )
    ui['old_project_open_btn'].click(
        fn=lock_project_page,
        inputs=object_count,
        outputs=[v for k, v in ui.items()]
    )
    # ui['old_project_open_btn'].click(
    #     fn=lock_project_page,
    #     inputs=object_count,
    #     outputs=[v for k, v in ui.items()]
    # )
    ui['old_project_drop'].focus(
        fn=list_project,
        inputs=ui['storage_path'],
        outputs=ui['old_project_drop']
    )
    ui['old_project_drop'].select(
        fn=unlock_old_project_btn,
        inputs=old_project_btn_count,
        outputs=old_project_btn_list,
    )
    ui['new_project_create_btn'].click(
        fn=create_new_project,
        inputs=[ui[it] for it in ['storage_path', 'new_project_name']]
    )
    
    ui['new_project_name'].focus(
        fn=default_new_project_name,
        inputs=ui['new_project_name'],
        outputs=ui['new_project_name']
    )
    ui['new_project_create_btn'].click(
        fn=set_project_name,
        inputs=ui['new_project_name'],
        outputs=ui['project_name']
    )
    ui['old_project_open_btn'].click(
        fn=set_project_name,
        inputs=ui['old_project_drop'],
        outputs=ui['project_name']
    )
    ui['old_project_delete_btn'].click(
        fn=delete_old_project,
        inputs=[ui['storage_path'], ui['old_project_drop']],
    )

    return ui



import cv2
import numpy as np

    
def get_centroids(frame, roi):
    connected_components = roi_connected_components(frame, roi)
    num_labels, _, stats, centroids = connected_components
    areas = [stats[i, cv2.CC_STAT_AREA] for i in range(1, num_labels)]
    max_label = np.argmax(areas)
    x = centroids[max_label + 1][0]
    y = centroids[max_label + 1][1]
    return x, y


def center_roi(frame, mask, roi_color):
    h, w = frame.shape[:2]
    center = (w // 2, h // 2)
    roi_x, roi_y = get_centroids(mask, roi_color)
    matrix = np.float32([[1, 0, center[0] - roi_x], [0, 1, center[1] - roi_y]])
    return cv2.warpAffine(frame, matrix, (w, h))



def rotate_based_on_roi_closest_center_point(frame, mask, roi_color):
    roi_contour = get_contour(mask, roi_color)
    h, w = frame.shape[:2]
    center_x, conter_y = (w // 2, h // 2)
    (closest_point_x, closest_point_y), _ = find_closest_point((center_x, conter_y), roi_contour)
    theta = np.arctan2(closest_point_y - conter_y, closest_point_x - center_x) * 180. / np.pi
    matrix = cv2.getRotationMatrix2D((center_x, conter_y), theta-90, 1.0)
    return cv2.warpAffine(frame, matrix, (w, h))



def crop(frame, crop_h, crop_w):
    h, w = frame.shape[:2]
    u, d = int(h // 2 - crop_h // 2), int(h // 2 + crop_h // 2)
    l, r = int(w // 2 - crop_w // 2), int(w // 2 + crop_w // 2)
    return frame[u:d, l:r]





def roi_connected_components(frame, select_roi, tolerance=30):

    if type(select_roi) == list and len(select_roi) == 3:
        h, w = frame.shape[:2]
        mask = np.zeros((h, w)).astype(np.uint8)
        lower_bound = np.array([select_roi[0] - tolerance,
                                select_roi[1] - tolerance,
                                select_roi[2] - tolerance])
        upper_bound = np.array([select_roi[0] + tolerance,
                                select_roi[1] + tolerance,
                                select_roi[2] + tolerance])
        within_range = cv2.inRange(frame, lower_bound, upper_bound)
        mask[within_range > 0] = 255
    else:
        select_roi = int(select_roi)

        mask = cv2.inRange(frame, select_roi, select_roi)


    
    output = cv2.connectedComponentsWithStats(mask, 8, cv2.CV_32S)
    num_labels = output[0]
    assert num_labels > 1, 'roi_connected_components error'
    return output



def get_contour(frame, roi):
    connected_components = roi_connected_components(frame, roi)
    num_labels, labels, stats, centroids = connected_components
    
    areas = [stats[j, cv2.CC_STAT_AREA] for j in range(1, num_labels)]
    maxi_comp_id = np.argmax(areas)
    selected_label = (labels == (maxi_comp_id+1)).astype(np.uint8) * 255
    _, binary_mask = cv2.threshold(selected_label, 0, 255, cv2.THRESH_BINARY)
    contour = cv2.findContours(binary_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)[0][0]
    contour = contour.squeeze()
    assert contour.ndim > 1, 'get_contour error'
    return contour
    



def get_mask(frame, select_roi):
    connected_components = roi_connected_components(frame, select_roi)
    num_labels, labels, stats, centroids = connected_components
    areas = [stats[i, cv2.CC_STAT_AREA] for i in range(1, num_labels)]
    max_label = np.argmax(areas)
    return (labels == (max_label+1))
    
    
def find_closest_point(ref, contour):
    mini = int(1e6)
    point_close = None
    assert len(contour) >= 2, 'find_closest_point error'
    for i in range(len(contour)):
        distance = cv2.norm(ref - contour[i])
        if distance < mini:
            mini = distance
            point_close = contour[i]

    return point_close, mini
    

def blank_page(crop_h, crop_w):
    white_frame = np.full((crop_h, crop_w, 3), 255, dtype=np.uint8)
    return white_frame
    



import random
import cv2
import numpy as np
from PIL import Image

import torch
import torchvision.transforms as TF
# import image_transforms as IT

cv2.setNumThreads(0)


class Resize(object):
    """Rescale the image in a sample to a given size.

    Args:
        output_size (tuple or int): Desired output size. If tuple, output is
            matched to output_size. If int, smaller of image edges is matched
            to output_size keeping aspect ratio the same.
    """
    def __init__(self, output_size, use_padding=False):
        assert isinstance(output_size, (int, tuple))
        if isinstance(output_size, int):
            self.output_size = (output_size, output_size)
        else:
            self.output_size = output_size
        self.use_padding = use_padding

    def __call__(self, sample):
        return self.padding(sample) if self.use_padding else self.rescale(
            sample)

    def rescale(self, sample):
        prev_img = sample['prev_img']
        h, w = prev_img.shape[:2]
        if self.output_size == (h, w):
            return sample
        else:
            new_h, new_w = self.output_size

        for elem in sample.keys():
            if 'meta' in elem:
                continue
            tmp = sample[elem]

            if elem == 'prev_img' or elem == 'curr_img' or elem == 'ref_img':
                flagval = cv2.INTER_CUBIC
            else:
                flagval = cv2.INTER_NEAREST

            if elem == 'curr_img' or elem == 'curr_label':
                new_tmp = []
                all_tmp = tmp
                for tmp in all_tmp:
                    tmp = cv2.resize(tmp,
                                     dsize=(new_w, new_h),
                                     interpolation=flagval)
                    new_tmp.append(tmp)
                tmp = new_tmp
            else:
                tmp = cv2.resize(tmp,
                                 dsize=(new_w, new_h),
                                 interpolation=flagval)

            sample[elem] = tmp

        return sample

    def padding(self, sample):
        prev_img = sample['prev_img']
        h, w = prev_img.shape[:2]
        if self.output_size == (h, w):
            return sample
        else:
            new_h, new_w = self.output_size

        def sep_pad(x):
            x0 = np.random.randint(0, x + 1)
            x1 = x - x0
            return x0, x1

        top_pad, bottom_pad = sep_pad(new_h - h)
        left_pad, right_pad = sep_pad(new_w - w)

        for elem in sample.keys():
            if 'meta' in elem:
                continue
            tmp = sample[elem]

            if elem == 'prev_img' or elem == 'curr_img' or elem == 'ref_img':
                pad_value = (124, 116, 104)
            else:
                pad_value = (0)

            if elem == 'curr_img' or elem == 'curr_label':
                new_tmp = []
                all_tmp = tmp
                for tmp in all_tmp:
                    tmp = cv2.copyMakeBorder(tmp,
                                             top_pad,
                                             bottom_pad,
                                             left_pad,
                                             right_pad,
                                             cv2.BORDER_CONSTANT,
                                             value=pad_value)
                    new_tmp.append(tmp)
                tmp = new_tmp
            else:
                tmp = cv2.copyMakeBorder(tmp,
                                         top_pad,
                                         bottom_pad,
                                         left_pad,
                                         right_pad,
                                         cv2.BORDER_CONSTANT,
                                         value=pad_value)

            sample[elem] = tmp

        return sample


class BalancedRandomCrop(object):
    """Crop randomly the image in a sample.

    Args:
        output_size (tuple or int): Desired output size. If int, square crop
            is made.
    """
    def __init__(self,
                 output_size,
                 max_step=5,
                 max_obj_num=5,
                 min_obj_pixel_num=100):
        assert isinstance(output_size, (int, tuple))
        if isinstance(output_size, int):
            self.output_size = (output_size, output_size)
        else:
            assert len(output_size) == 2
            self.output_size = output_size
        self.max_step = max_step
        self.max_obj_num = max_obj_num
        self.min_obj_pixel_num = min_obj_pixel_num

    def __call__(self, sample):

        image = sample['prev_img']
        h, w = image.shape[:2]
        new_h, new_w = self.output_size
        new_h = h if new_h >= h else new_h
        new_w = w if new_w >= w else new_w
        ref_label = sample["ref_label"]
        prev_label = sample["prev_label"]
        curr_label = sample["curr_label"]

        is_contain_obj = False
        step = 0
        while (not is_contain_obj) and (step < self.max_step):
            step += 1
            top = np.random.randint(0, h - new_h + 1)
            left = np.random.randint(0, w - new_w + 1)
            after_crop = []
            contains = []
            for elem in ([ref_label, prev_label] + curr_label):
                tmp = elem[top:top + new_h, left:left + new_w]
                contains.append(np.unique(tmp))
                after_crop.append(tmp)

            all_obj = list(np.sort(contains[0]))

            if all_obj[-1] == 0:
                continue

            # remove background
            if all_obj[0] == 0:
                all_obj = all_obj[1:]

            # remove small obj
            new_all_obj = []
            for obj_id in all_obj:
                after_crop_pixels = np.sum(after_crop[0] == obj_id)
                if after_crop_pixels > self.min_obj_pixel_num:
                    new_all_obj.append(obj_id)

            if len(new_all_obj) == 0:
                is_contain_obj = False
            else:
                is_contain_obj = True

            if len(new_all_obj) > self.max_obj_num:
                random.shuffle(new_all_obj)
                new_all_obj = new_all_obj[:self.max_obj_num]

            all_obj = [0] + new_all_obj

        post_process = []
        for elem in after_crop:
            new_elem = elem * 0
            for idx in range(len(all_obj)):
                obj_id = all_obj[idx]
                if obj_id == 0:
                    continue
                mask = elem == obj_id

                new_elem += (mask * idx).astype(np.uint8)
            post_process.append(new_elem.astype(np.uint8))

        sample["ref_label"] = post_process[0]
        sample["prev_label"] = post_process[1]
        curr_len = len(sample["curr_img"])
        sample["curr_label"] = []
        for idx in range(curr_len):
            sample["curr_label"].append(post_process[idx + 2])

        for elem in sample.keys():
            if 'meta' in elem or 'label' in elem:
                continue
            if elem == 'curr_img':
                new_tmp = []
                for tmp_ in sample[elem]:
                    tmp_ = tmp_[top:top + new_h, left:left + new_w]
                    new_tmp.append(tmp_)
                sample[elem] = new_tmp
            else:
                tmp = sample[elem]
                tmp = tmp[top:top + new_h, left:left + new_w]
                sample[elem] = tmp

        obj_num = len(all_obj) - 1

        sample['meta']['obj_num'] = obj_num

        return sample


class RandomScale(object):
    """Randomly resize the image and the ground truth to specified scales.
    Args:
        scales (list): the list of scales
    """
    def __init__(self, min_scale=1., max_scale=1.3, short_edge=None):
        self.min_scale = min_scale
        self.max_scale = max_scale
        self.short_edge = short_edge

    def __call__(self, sample):
        # Fixed range of scales
        sc = np.random.uniform(self.min_scale, self.max_scale)
        # Align short edge
        if self.short_edge is not None:
            image = sample['prev_img']
            h, w = image.shape[:2]
            if h > w:
                sc *= float(self.short_edge) / w
            else:
                sc *= float(self.short_edge) / h

        for elem in sample.keys():
            if 'meta' in elem:
                continue
            tmp = sample[elem]

            if elem == 'prev_img' or elem == 'curr_img' or elem == 'ref_img':
                flagval = cv2.INTER_CUBIC
            else:
                flagval = cv2.INTER_NEAREST

            if elem == 'curr_img' or elem == 'curr_label':
                new_tmp = []
                for tmp_ in tmp:
                    tmp_ = cv2.resize(tmp_,
                                      None,
                                      fx=sc,
                                      fy=sc,
                                      interpolation=flagval)
                    new_tmp.append(tmp_)
                tmp = new_tmp
            else:
                tmp = cv2.resize(tmp,
                                 None,
                                 fx=sc,
                                 fy=sc,
                                 interpolation=flagval)

            sample[elem] = tmp

        return sample


class RandomScaleV2(object):
    """Randomly resize the image and the ground truth to specified scales.
    Args:
        scales (list): the list of scales
    """
    def __init__(self,
                 min_scale=0.36,
                 max_scale=1.0,
                 short_edge=None,
                 ratio=[3. / 4., 4. / 3.]):
        self.min_scale = min_scale
        self.max_scale = max_scale
        self.short_edge = short_edge
        self.ratio = ratio

    def __call__(self, sample):
        image = sample['prev_img']
        h, w = image.shape[:2]

        new_h, new_w = self.get_params(h, w)

        sc_x = float(new_w) / w
        sc_y = float(new_h) / h

        # Align short edge
        if not (self.short_edge is None):
            if h > w:
                sc_x *= float(self.short_edge) / w
                sc_y *= float(self.short_edge) / w
            else:
                sc_x *= float(self.short_edge) / h
                sc_y *= float(self.short_edge) / h

        for elem in sample.keys():
            if 'meta' in elem:
                continue
            tmp = sample[elem]

            if elem == 'prev_img' or elem == 'curr_img' or elem == 'ref_img':
                flagval = cv2.INTER_CUBIC
            else:
                flagval = cv2.INTER_NEAREST

            if elem == 'curr_img' or elem == 'curr_label':
                new_tmp = []
                for tmp_ in tmp:
                    tmp_ = cv2.resize(tmp_,
                                      None,
                                      fx=sc_x,
                                      fy=sc_y,
                                      interpolation=flagval)
                    new_tmp.append(tmp_)
                tmp = new_tmp
            else:
                tmp = cv2.resize(tmp,
                                 None,
                                 fx=sc_x,
                                 fy=sc_y,
                                 interpolation=flagval)

            sample[elem] = tmp

        return sample

    def get_params(self, height, width):
        area = height * width

        log_ratio = [np.log(item) for item in self.ratio]
        for _ in range(10):
            target_area = area * np.random.uniform(self.min_scale**2,
                                                   self.max_scale**2)
            aspect_ratio = np.exp(np.random.uniform(log_ratio[0],
                                                    log_ratio[1]))

            w = int(round(np.sqrt(target_area * aspect_ratio)))
            h = int(round(np.sqrt(target_area / aspect_ratio)))

            if 0 < w <= width and 0 < h <= height:
                return h, w

        # Fallback to central crop
        in_ratio = float(width) / float(height)
        if in_ratio < min(self.ratio):
            w = width
            h = int(round(w / min(self.ratio)))
        elif in_ratio > max(self.ratio):
            h = height
            w = int(round(h * max(self.ratio)))
        else:  # whole image
            w = width
            h = height

        return h, w

class RestrictSize(object):
    """Randomly resize the image and the ground truth to specified scales.
    Args:
        scales (list): the list of scales
    """
    def __init__(self, max_short_edge=None, max_long_edge=800 * 1.3):
        self.max_short_edge = max_short_edge
        self.max_long_edge = max_long_edge
        assert ((max_short_edge is None)) or ((max_long_edge is None))

    def __call__(self, sample):

        # Fixed range of scales
        sc = None
        image = sample['ref_img']
        h, w = image.shape[:2]
        # Align short edge
        if not (self.max_short_edge is None):
            if h > w:
                short_edge = w
            else:
                short_edge = h
            if short_edge < self.max_short_edge:
                sc = float(self.max_short_edge) / short_edge
        else:
            if h > w:
                long_edge = h
            else:
                long_edge = w
            if long_edge > self.max_long_edge:
                sc = float(self.max_long_edge) / long_edge

        if sc is None:
            new_h = h
            new_w = w
        else:
            new_h = int(sc * h)
            new_w = int(sc * w)
        new_h = new_h - (new_h - 1) % 4
        new_w = new_w - (new_w - 1) % 4
        if new_h == h and new_w == w:
            return sample

        for elem in sample.keys():
            if 'meta' in elem:
                continue
            tmp = sample[elem]

            if 'label' in elem:
                flagval = cv2.INTER_NEAREST
            else:
                flagval = cv2.INTER_CUBIC

            tmp = cv2.resize(tmp, dsize=(new_w, new_h), interpolation=flagval)

            sample[elem] = tmp

        return sample


class RandomHorizontalFlip(object):
    """Horizontally flip the given image and ground truth randomly with a probability of 0.5."""
    def __init__(self, prob):
        self.p = prob

    def __call__(self, sample):

        if random.random() < self.p:
            for elem in sample.keys():
                if 'meta' in elem:
                    continue
                if elem == 'curr_img' or elem == 'curr_label':
                    new_tmp = []
                    for tmp_ in sample[elem]:
                        tmp_ = cv2.flip(tmp_, flipCode=1)
                        new_tmp.append(tmp_)
                    sample[elem] = new_tmp
                else:
                    tmp = sample[elem]
                    tmp = cv2.flip(tmp, flipCode=1)
                    sample[elem] = tmp

        return sample


class RandomVerticalFlip(object):
    """Vertically flip the given image and ground truth randomly with a probability of 0.5."""
    def __init__(self, prob=0.3):
        self.p = prob

    def __call__(self, sample):

        if random.random() < self.p:
            for elem in sample.keys():
                if 'meta' in elem:
                    continue
                if elem == 'curr_img' or elem == 'curr_label':
                    new_tmp = []
                    for tmp_ in sample[elem]:
                        tmp_ = cv2.flip(tmp_, flipCode=0)
                        new_tmp.append(tmp_)
                    sample[elem] = new_tmp
                else:
                    tmp = sample[elem]
                    tmp = cv2.flip(tmp, flipCode=0)
                    sample[elem] = tmp

        return sample


class RandomGaussianBlur(object):
    def __init__(self, prob=0.3, sigma=[0.1, 2.]):
        self.aug = TF.RandomApply([TF.GaussianBlur(sigma)], p=prob)

    def __call__(self, sample):
        for elem in sample.keys():
            if 'meta' in elem or 'label' in elem:
                continue

            if elem == 'curr_img':
                new_tmp = []
                for tmp_ in sample[elem]:
                    tmp_ = self.apply_augmentation(tmp_)
                    new_tmp.append(tmp_)
                sample[elem] = new_tmp
            else:
                tmp = sample[elem]
                tmp = self.apply_augmentation(tmp)
                sample[elem] = tmp
        return sample

    def apply_augmentation(self, x):
        x = Image.fromarray(np.uint8(x))
        x = self.aug(x)
        x = np.array(x, dtype=np.float32)
        return x


class RandomGrayScale(RandomGaussianBlur):
    def __init__(self, prob=0.2):
        self.aug = TF.RandomGrayscale(p=prob)


class RandomColorJitter(RandomGaussianBlur):
    def __init__(self,
                 prob=0.8,
                 brightness=0.4,
                 contrast=0.4,
                 saturation=0.2,
                 hue=0.1):
        self.aug = TF.RandomApply(
            [TF.ColorJitter(brightness, contrast, saturation, hue)], p=prob)


class SubtractMeanImage(object):
    def __init__(self, mean, change_channels=False):
        self.mean = mean
        self.change_channels = change_channels

    def __call__(self, sample):
        for elem in sample.keys():
            if 'image' in elem:
                if self.change_channels:
                    sample[elem] = sample[elem][:, :, [2, 1, 0]]
                sample[elem] = np.subtract(
                    sample[elem], np.array(self.mean, dtype=np.float32))
        return sample

    def __str__(self):
        return 'SubtractMeanImage' + str(self.mean)


class ToTensor(object):
    """Convert ndarrays in sample to Tensors."""
    def __call__(self, sample):

        for elem in sample.keys():
            if 'meta' in elem:
                continue
            tmp = sample[elem]

            if elem == 'curr_img' or elem == 'curr_label':
                new_tmp = []
                for tmp_ in tmp:
                    if tmp_.ndim == 2:
                        tmp_ = tmp_[:, :, np.newaxis]
                        tmp_ = tmp_.transpose((2, 0, 1))
                        new_tmp.append(torch.from_numpy(tmp_).int())
                    else:
                        tmp_ = tmp_ / 255.
                        tmp_ -= (0.485, 0.456, 0.406)
                        tmp_ /= (0.229, 0.224, 0.225)
                        tmp_ = tmp_.transpose((2, 0, 1))
                        new_tmp.append(torch.from_numpy(tmp_))
                tmp = new_tmp
            else:
                if tmp.ndim == 2:
                    tmp = tmp[:, :, np.newaxis]
                    tmp = tmp.transpose((2, 0, 1))
                    tmp = torch.from_numpy(tmp).int()
                else:
                    tmp = tmp / 255.
                    tmp -= (0.485, 0.456, 0.406)
                    tmp /= (0.229, 0.224, 0.225)
                    tmp = tmp.transpose((2, 0, 1))
                    tmp = torch.from_numpy(tmp)
            sample[elem] = tmp

        return sample


class MultiRestrictSize(object):
    def __init__(self,
                 max_short_edge=None,
                 max_long_edge=800,
                 flip=False,
                 multi_scale=[1.3],
                 align_corners=True,
                 max_stride=16):
        self.max_short_edge = max_short_edge
        self.max_long_edge = max_long_edge
        self.multi_scale = multi_scale
        self.flip = flip
        self.align_corners = align_corners
        self.max_stride = max_stride

    def __call__(self, sample):
        samples = []
        image = sample['current_img']
        h, w = image.shape[:2]
        for scale in self.multi_scale:
            # restrict short edge
            sc = 1.
            if self.max_short_edge is not None:
                if h > w:
                    short_edge = w
                else:
                    short_edge = h
                if short_edge > self.max_short_edge:
                    sc *= float(self.max_short_edge) / short_edge
            new_h, new_w = sc * h, sc * w

            # restrict long edge
            sc = 1.
            if self.max_long_edge is not None:
                if new_h > new_w:
                    long_edge = new_h
                else:
                    long_edge = new_w
                if long_edge > self.max_long_edge:
                    sc *= float(self.max_long_edge) / long_edge

            new_h, new_w = sc * new_h, sc * new_w

            new_h = int(new_h * scale)
            new_w = int(new_w * scale)

            if self.align_corners:
                if (new_h - 1) % self.max_stride != 0:
                    new_h = int(
                        np.around((new_h - 1) / self.max_stride) *
                        self.max_stride + 1)
                if (new_w - 1) % self.max_stride != 0:
                    new_w = int(
                        np.around((new_w - 1) / self.max_stride) *
                        self.max_stride + 1)
            else:
                if new_h % self.max_stride != 0:
                    new_h = int(
                        np.around(new_h / self.max_stride) * self.max_stride)
                if new_w % self.max_stride != 0:
                    new_w = int(
                        np.around(new_w / self.max_stride) * self.max_stride)

            if new_h == h and new_w == w:
                samples.append(sample)
            else:
                new_sample = {}
                for elem in sample.keys():
                    if 'meta' in elem:
                        new_sample[elem] = sample[elem]
                        continue
                    tmp = sample[elem]
                    if 'label' in elem:
                        new_sample[elem] = sample[elem]
                        continue
                    else:
                        flagval = cv2.INTER_CUBIC
                        tmp = cv2.resize(tmp,
                                         dsize=(new_w, new_h),
                                         interpolation=flagval)
                        new_sample[elem] = tmp
                samples.append(new_sample)

            if self.flip:
                now_sample = samples[-1]
                new_sample = {}
                for elem in now_sample.keys():
                    if 'meta' in elem:
                        new_sample[elem] = now_sample[elem].copy()
                        new_sample[elem]['flip'] = True
                        continue
                    tmp = now_sample[elem]
                    tmp = tmp[:, ::-1].copy()
                    new_sample[elem] = tmp
                samples.append(new_sample)

        return samples


class MultiToTensor(object):
    def __call__(self, samples):
        for idx in range(len(samples)):
            sample = samples[idx]
            for elem in sample.keys():
                if 'meta' in elem:
                    continue
                tmp = sample[elem]
                if tmp is None:
                    continue

                if tmp.ndim == 2:
                    tmp = tmp[:, :, np.newaxis]
                    tmp = tmp.transpose((2, 0, 1))
                    samples[idx][elem] = torch.from_numpy(tmp).int()
                else:
                    tmp = tmp / 255.
                    tmp -= (0.485, 0.456, 0.406)
                    tmp /= (0.229, 0.224, 0.225)
                    tmp = tmp.transpose((2, 0, 1))
                    samples[idx][elem] = torch.from_numpy(tmp)

        return samples



import torch


def pytorch_iou(pred, target, obj_num, epsilon=1e-6):
    '''
    pred: [bs, h, w]
    target: [bs, h, w]
    obj_num: [bs]
    '''
    bs = pred.size(0)
    all_iou = []
    for idx in range(bs):
        now_pred = pred[idx].unsqueeze(0)
        now_target = target[idx].unsqueeze(0)
        now_obj_num = obj_num[idx]

        obj_ids = torch.arange(0, now_obj_num + 1,
                               device=now_pred.device).int().view(-1, 1, 1)
        if obj_ids.size(0) == 1:  # only contain background
            continue
        else:
            obj_ids = obj_ids[1:]
            now_pred = (now_pred == obj_ids).float()
            now_target = (now_target == obj_ids).float()

            intersection = (now_pred * now_target).sum((1, 2))
            union = ((now_pred + now_target) > 0).float().sum((1, 2))

            now_iou = (intersection + epsilon) / (union + epsilon)

            all_iou.append(now_iou.mean())
    if len(all_iou) > 0:
        all_iou = torch.stack(all_iou).mean()
    else:
        all_iou = torch.ones((1), device=pred.device)
    return all_iou



import os
import importlib


class DefaultEngineConfig():
    def __init__(self, exp_name='default', model='aott'):
        model_cfg = importlib.import_module(f'castle.aot.configs.models.{model}').ModelConfig()
        self.__dict__.update(model_cfg.__dict__)  # add model config

        self.EXP_NAME = exp_name + '_' + self.MODEL_NAME

        self.STAGE_NAME = 'YTB'

        self.DATASETS = ['youtubevos']
        self.DATA_WORKERS = 8
        self.DATA_RANDOMCROP = (465,
                                465) if self.MODEL_ALIGN_CORNERS else (464,
                                                                       464)
        self.DATA_RANDOMFLIP = 0.5
        self.DATA_MAX_CROP_STEPS = 10
        self.DATA_SHORT_EDGE_LEN = 480
        self.DATA_MIN_SCALE_FACTOR = 0.7
        self.DATA_MAX_SCALE_FACTOR = 1.3
        self.DATA_RANDOM_REVERSE_SEQ = True
        self.DATA_SEQ_LEN = 5
        self.DATA_DAVIS_REPEAT = 5
        self.DATA_RANDOM_GAP_DAVIS = 12  # max frame interval between two sampled frames for DAVIS (24fps)
        self.DATA_RANDOM_GAP_YTB = 3  # max frame interval between two sampled frames for YouTube-VOS (6fps)
        self.DATA_DYNAMIC_MERGE_PROB = 0.3

        self.PRETRAIN = True
        self.PRETRAIN_FULL = False  # if False, load encoder only
        self.PRETRAIN_MODEL = './data_wd/pretrain_model/mobilenet_v2.pth'
        # self.PRETRAIN_MODEL = './pretrain_models/mobilenet_v2-b0353104.pth'

        self.TRAIN_TOTAL_STEPS = 100000
        self.TRAIN_START_STEP = 0
        self.TRAIN_WEIGHT_DECAY = 0.07
        self.TRAIN_WEIGHT_DECAY_EXCLUSIVE = {
            # 'encoder.': 0.01
        }
        self.TRAIN_WEIGHT_DECAY_EXEMPTION = [
            'absolute_pos_embed', 'relative_position_bias_table',
            'relative_emb_v', 'conv_out'
        ]
        self.TRAIN_LR = 2e-4
        self.TRAIN_LR_MIN = 2e-5 if 'mobilenetv2' in self.MODEL_ENCODER else 1e-5
        self.TRAIN_LR_POWER = 0.9
        self.TRAIN_LR_ENCODER_RATIO = 0.1
        self.TRAIN_LR_WARM_UP_RATIO = 0.05
        self.TRAIN_LR_COSINE_DECAY = False
        self.TRAIN_LR_RESTART = 1
        self.TRAIN_LR_UPDATE_STEP = 1
        self.TRAIN_AUX_LOSS_WEIGHT = 1.0
        self.TRAIN_AUX_LOSS_RATIO = 1.0
        self.TRAIN_OPT = 'adamw'
        self.TRAIN_SGD_MOMENTUM = 0.9
        self.TRAIN_GPUS = 4
        self.TRAIN_BATCH_SIZE = 16
        self.TRAIN_TBLOG = False
        self.TRAIN_TBLOG_STEP = 50
        self.TRAIN_LOG_STEP = 20
        self.TRAIN_IMG_LOG = True
        self.TRAIN_TOP_K_PERCENT_PIXELS = 0.15
        self.TRAIN_SEQ_TRAINING_FREEZE_PARAMS = ['patch_wise_id_bank']
        self.TRAIN_SEQ_TRAINING_START_RATIO = 0.5
        self.TRAIN_HARD_MINING_RATIO = 0.5
        self.TRAIN_EMA_RATIO = 0.1
        self.TRAIN_CLIP_GRAD_NORM = 5.
        self.TRAIN_SAVE_STEP = 5000
        self.TRAIN_MAX_KEEP_CKPT = 8
        self.TRAIN_RESUME = False
        self.TRAIN_RESUME_CKPT = None
        self.TRAIN_RESUME_STEP = 0
        self.TRAIN_AUTO_RESUME = True
        self.TRAIN_DATASET_FULL_RESOLUTION = False
        self.TRAIN_ENABLE_PREV_FRAME = False
        self.TRAIN_ENCODER_FREEZE_AT = 2
        self.TRAIN_LSTT_EMB_DROPOUT = 0.
        self.TRAIN_LSTT_ID_DROPOUT = 0.
        self.TRAIN_LSTT_DROPPATH = 0.1
        self.TRAIN_LSTT_DROPPATH_SCALING = False
        self.TRAIN_LSTT_DROPPATH_LST = False
        self.TRAIN_LSTT_LT_DROPOUT = 0.
        self.TRAIN_LSTT_ST_DROPOUT = 0.

        self.TEST_GPU_ID = 0
        self.TEST_GPU_NUM = 1
        self.TEST_FRAME_LOG = False
        self.TEST_DATASET = 'youtubevos'
        self.TEST_DATASET_FULL_RESOLUTION = False
        self.TEST_DATASET_SPLIT = 'val'
        self.TEST_CKPT_PATH = None
        # if "None", evaluate the latest checkpoint.
        self.TEST_CKPT_STEP = None
        self.TEST_FLIP = False
        self.TEST_MULTISCALE = [1]
        self.TEST_MAX_SHORT_EDGE = None
        self.TEST_MAX_LONG_EDGE = 800 * 1.3
        self.TEST_WORKERS = 4

        # GPU distribution
        self.DIST_ENABLE = True
        self.DIST_BACKEND = "nccl"  # "gloo"
        # self.DIST_URL = "tcp://127.0.0.1:13241"
        self.DIST_START_GPU = 0

    def init_dir(self):
        self.DIR_DATA = '../VOS02/datasets'#'./datasets'
        self.DIR_DAVIS = os.path.join(self.DIR_DATA, 'DAVIS')
        self.DIR_YTB = os.path.join(self.DIR_DATA, 'YTB')
        self.DIR_STATIC = os.path.join(self.DIR_DATA, 'Static')

        self.DIR_ROOT = './'#'./data_wd/youtube_vos_jobs'

        self.DIR_RESULT = os.path.join(self.DIR_ROOT, 'result', self.EXP_NAME,
                                       self.STAGE_NAME)
        self.DIR_CKPT = os.path.join(self.DIR_RESULT, 'ckpt')
        self.DIR_EMA_CKPT = os.path.join(self.DIR_RESULT, 'ema_ckpt')
        self.DIR_LOG = os.path.join(self.DIR_RESULT, 'log')
        self.DIR_TB_LOG = os.path.join(self.DIR_RESULT, 'log', 'tensorboard')
        self.DIR_IMG_LOG = os.path.join(self.DIR_RESULT, 'log', 'img')
        self.DIR_EVALUATION = os.path.join(self.DIR_RESULT, 'eval')
        self.DIR_IMG_LOG = './img_logs'
        self.DIR_EVALUATION = './results'

        # for path in [
        #         self.DIR_RESULT, self.DIR_CKPT, self.DIR_EMA_CKPT,
        #         self.DIR_LOG, self.DIR_EVALUATION, self.DIR_IMG_LOG,
        #         self.DIR_TB_LOG
        # ]:
        #     if not os.path.isdir(path):
        #         try:
        #             os.makedirs(path)
        #         except Exception as inst:
        #             print(inst)
        #             print('Failed to make dir: {}.'.format(path))



import gradio as gr
import os
import json
import glob
import numpy as np

def init_drop(storage_path, project_name):
    if project_name == None:
        return gr.update(choices=[])

    project_path = os.path.join(storage_path, project_name)
    project_config_path = os.path.join(project_path, 'config.json')
    project_config = json.load(open(project_config_path, 'r'))
    latent_list = [k for k, v in project_config['latent'].items()]
    latent_list.append("All")


    return gr.update(choices=latent_list), gr.update(choices=[]), latent_list, []



def select_latent(target_latent,  unselect_list, select_list):
    if target_latent == 'All':
        unselect_list.remove('All')
        select_list.extend(unselect_list)
        unselect_list = ['All']
    else:
        unselect_list.remove(target_latent)
        select_list.append(target_latent)
    return gr.update(choices=unselect_list), gr.update(choices=select_list), unselect_list, select_list

def unselect_latent(target_latent,  unselect_list, select_list):
    select_list.remove(target_latent)
    unselect_list.append(target_latent)
    return gr.update(choices=unselect_list), gr.update(choices=select_list, value=""), unselect_list, select_list


def build_umap_intput(storage_path, project_name, umap_input_name, select_list):
    if len(umap_input_name) == 0:
        gr.Info("The name is not allow empty.")
        return
    project_path = os.path.join(storage_path, project_name)
    project_config_path = os.path.join(project_path, 'config.json')
    project_config = json.load(open(project_config_path, 'r'))
    latent_dir_path = os.path.join(storage_path, project_name, 'latent')
    if not 'umap_input' in project_config:
        project_config['umap_input'] = dict()

    umap_input_dir_path = os.path.join(project_path, 'umap_input')
    os.makedirs(umap_input_dir_path, exist_ok=True)
    umap_input_path = os.path.join(umap_input_dir_path, f'{umap_input_name}_umap_input.npz')

    umap_input_mask = []
    for it in select_list:
        latent = np.load(f'{latent_dir_path}/{it}')['latent']
        mask = (~np.isnan(latent[:, 0]))
        umap_input_mask.extend(mask)

    umap_input_mask = np.array(umap_input_mask)
    # print('umap_input_mask', umap_input_mask.shape)

    np.savez_compressed(umap_input_path, umap_mask=umap_input_mask)
    project_config['umap_input'][f'{umap_input_name}_umap_input.npz'] = {
        'latent': select_list,
        'source': [project_config['latent'][it] for it in select_list],

    }
    json.dump(project_config, open(project_config_path,'w'))
    gr.Info("Successful build the list as a UMAP Input")
    return



def create_cluster_input_ui(storage_path, project_name, cluster_page_tab):
    ui = dict()

    select_list = gr.State([])
    unselect_list = gr.State([])
    with gr.Row(visible=True):
        with gr.Column(scale=5):
            ui['unselect_drop'] = gr.Dropdown(label='Unselect List', interactive=True)
            ui['select_latent'] = gr.Button('Select')
        with gr.Column(scale=5):
            ui['select_drop'] = gr.Dropdown(label='Select List', interactive=True)
            ui['unselect_latent'] = gr.Button('Unselect')
    ui['name'] = gr.Textbox(label='Name')
    ui['build'] = gr.Button('Build List as UMAP Input')



    cluster_page_tab.select(
        fn=init_drop,
        inputs=[storage_path, project_name],
        outputs=[ui['unselect_drop'], ui['select_drop'], unselect_list, select_list]
    )


    ui['select_latent'].click(
        fn=select_latent,
        inputs=[ui['unselect_drop'], unselect_list, select_list],
        outputs=[ui['unselect_drop'], ui['select_drop'], unselect_list, select_list],
    )

    ui['unselect_latent'].click(
        fn=unselect_latent,
        inputs=[ui['select_drop'], unselect_list, select_list],
        outputs=[ui['unselect_drop'], ui['select_drop'], unselect_list, select_list],
    )

    ui['build'].click(
        fn=build_umap_intput,
        inputs=[storage_path, project_name, ui['name'], select_list],
    )


    return ui


import numpy as np

from ...utils.image import one_hot_mask

from ..layers.basic import seq_to_2d
from .aot_engine import AOTEngine, AOTInferEngine


class DeAOTEngine(AOTEngine):
    def __init__(self,
                 aot_model,
                 device,
                 long_term_mem_gap=9999,
                 short_term_mem_skip=1,
                 layer_loss_scaling_ratio=2.,
                 max_len_long_term=9999):
        super().__init__(aot_model, device, long_term_mem_gap,
                         short_term_mem_skip, max_len_long_term)
        self.layer_loss_scaling_ratio = layer_loss_scaling_ratio
    def update_short_term_memory(self, curr_mask, curr_id_emb=None, skip_long_term_update=False):

        if curr_id_emb is None:
            if len(curr_mask.size()) == 3 or curr_mask.size()[0] == 1:
                curr_one_hot_mask = one_hot_mask(curr_mask, self.max_obj_num)
            else:
                curr_one_hot_mask = curr_mask
            curr_id_emb = self.assign_identity(curr_one_hot_mask)

        lstt_curr_memories = self.curr_lstt_output[1]
        lstt_curr_memories_2d = []
        for layer_idx in range(len(lstt_curr_memories)):
            curr_k, curr_v, curr_id_k, curr_id_v = lstt_curr_memories[
                layer_idx]
            curr_id_k, curr_id_v = self.AOT.LSTT.layers[
                layer_idx].fuse_key_value_id(curr_id_k, curr_id_v, curr_id_emb)
            lstt_curr_memories[layer_idx][2], lstt_curr_memories[layer_idx][
                3] = curr_id_k, curr_id_v
            local_curr_id_k = seq_to_2d(
                curr_id_k, self.enc_size_2d) if curr_id_k is not None else None
            local_curr_id_v = seq_to_2d(curr_id_v, self.enc_size_2d)
            lstt_curr_memories_2d.append([
                seq_to_2d(curr_k, self.enc_size_2d),
                seq_to_2d(curr_v, self.enc_size_2d), local_curr_id_k,
                local_curr_id_v
            ])

        self.short_term_memories_list.append(lstt_curr_memories_2d)
        self.short_term_memories_list = self.short_term_memories_list[
            -self.short_term_mem_skip:]
        self.short_term_memories = self.short_term_memories_list[0]

        if self.frame_step - self.last_mem_step >= self.long_term_mem_gap:
            # skip the update of long-term memory or not
            if not skip_long_term_update: 
                self.update_long_term_memory(lstt_curr_memories)
            self.last_mem_step = self.frame_step


class DeAOTInferEngine(AOTInferEngine):
    def __init__(self,
                 aot_model,
                 device,
                 long_term_mem_gap=9999,
                 short_term_mem_skip=1,
                 max_aot_obj_num=None,
                 max_len_long_term=9999):
        super().__init__(aot_model, device, long_term_mem_gap,
                         short_term_mem_skip, max_aot_obj_num, max_len_long_term)
    def add_reference_frame(self, img, mask, obj_nums, frame_step=-1):
        if isinstance(obj_nums, list):
            obj_nums = obj_nums[0]
        self.obj_nums = obj_nums
        aot_num = max(np.ceil(obj_nums / self.max_aot_obj_num), 1)
        while (aot_num > len(self.aot_engines)):
            new_engine = DeAOTEngine(self.AOT, self.device,
                                     self.long_term_mem_gap,
                                     self.short_term_mem_skip,
                                     max_len_long_term = self.max_len_long_term)
            new_engine.eval()
            self.aot_engines.append(new_engine)

        separated_masks, separated_obj_nums = self.separate_mask(
            mask, obj_nums)
        img_embs = None
        for aot_engine, separated_mask, separated_obj_num in zip(
                self.aot_engines, separated_masks, separated_obj_nums):
            # if aot_engine.obj_nums is None or aot_engine.obj_nums[0] < separated_obj_num:
            aot_engine.add_reference_frame(img,
                                        separated_mask,
                                        obj_nums=[separated_obj_num],
                                        frame_step=frame_step,
                                        img_embs=img_embs)
            # else:
                # aot_engine.update_short_term_memory(separated_mask)
            if img_embs is None:  # reuse image embeddings
                img_embs = aot_engine.curr_enc_embs
 
        self.update_size()

    def add_reference_frame_incremental(self, img, mask, obj_nums, frame_step=-1):
            if isinstance(obj_nums, list):
                obj_nums = obj_nums[0]
            self.obj_nums = obj_nums
            aot_num = max(np.ceil(obj_nums / self.max_aot_obj_num), 1)
            while (aot_num > len(self.aot_engines)):
                new_engine = DeAOTEngine(self.AOT, self.device,
                                    self.long_term_mem_gap,
                                    self.short_term_mem_skip)
                new_engine.eval()
                self.aot_engines.append(new_engine)

            separated_masks, separated_obj_nums = self.separate_mask(
                mask, obj_nums)
            img_embs = None
            for aot_engine, separated_mask, separated_obj_num in zip(
                    self.aot_engines, separated_masks, separated_obj_nums):
                if aot_engine.obj_nums is None or aot_engine.obj_nums[0] < separated_obj_num:
                    aot_engine.add_reference_frame(img,
                                                separated_mask,
                                                obj_nums=[separated_obj_num],
                                                frame_step=frame_step,
                                                img_embs=img_embs)
                else:
                    aot_engine.update_short_term_memory(separated_mask)
                    
                if img_embs is None:  # reuse image embeddings
                    img_embs = aot_engine.curr_enc_embs

            self.update_size()


from .default import DefaultModelConfig as BaseConfig


class DefaultModelConfig(BaseConfig):
    def __init__(self):
        super().__init__()
        self.MODEL_NAME = 'DeAOTDefault'

        self.MODEL_VOS = 'deaot'
        self.MODEL_ENGINE = 'deaotengine'

        self.MODEL_DECODER_INTERMEDIATE_LSTT = False

        self.MODEL_SELF_HEADS = 1
        self.MODEL_ATT_HEADS = 1

        self.TRAIN_AUG_TYPE = 'v2'



import gradio as gr


from .project_ui import create_project_ui
from .source_ui import create_source_ui
from .edit_ui import create_edit_ui
from .extract_ui import create_extract_ui
from .cluster_page_ui import create_cluster_page_ui

def show_ui(project_name, object_count):
    if not project_name == None:
        return [gr.update(visible=True) for i in range(object_count)]
    else:
        return [gr.update(visible=False) for i in range(object_count)]


def create_ui(OS_SYS, root=''):
    
    with gr.Blocks(theme=gr.themes.Soft()) as app:
        with gr.Tab(label='Project'):
            project_ui = create_project_ui(OS_SYS, root)
            project_name = project_ui['project_name']
            storage_path = project_ui['storage_path']

        
        with gr.Tab(label='Upload Videos') as source_tab:
            source_ui = create_source_ui(storage_path, project_name)
            source_ui_object_count = gr.State(len(source_ui))
            source_tab.select(
                fn=show_ui, 
                inputs=[project_ui['project_name'], source_ui_object_count], 
                outputs=[v for k, v in source_ui.items()]
            )

        with gr.Tab(label='Tracking ROIs') as edit_tab:
            edit_ui = create_edit_ui(storage_path, project_name)
            edit_ui_object_count = gr.State(len(edit_ui))
            edit_tab.select(
                fn=show_ui, 
                inputs=[project_ui['project_name'], edit_ui_object_count], 
                outputs=[v for k, v in edit_ui.items()]
            )

        with gr.Tab(label='Extract Latent') as extract_tab:
            extract_ui = create_extract_ui(storage_path, project_name, extract_tab)
            extract_ui_object_count = gr.State(len(extract_ui))
            extract_tab.select(
                fn=show_ui, 
                inputs=[project_ui['project_name'], extract_ui_object_count], 
                outputs=[v for k, v in extract_ui.items()]
            )
            pass

        with gr.Tab(label='Latent Explorer') as cluster_page_tab:
            cluster_ui = create_cluster_page_ui(storage_path, project_name, cluster_page_tab)
            cluster_ui_object_count = gr.State(len(cluster_ui))
            cluster_page_tab.select(
                fn=show_ui, 
                inputs=[project_ui['project_name'], cluster_ui_object_count], 
                outputs=[v for k, v in cluster_ui.items()]
            )
            pass

        # with gr.Tab(label='Export'):
        #     pass

    return app




import torch.nn as nn

from castle.aot.networks.layers.transformer import DualBranchGPM
from .aot import AOT
from castle.aot.networks.decoders import build_decoder


class DeAOT(AOT):
    def __init__(self, cfg, encoder='mobilenetv2', decoder='fpn'):
        super().__init__(cfg, encoder, decoder)

        self.LSTT = DualBranchGPM(
            cfg.MODEL_LSTT_NUM,
            cfg.MODEL_ENCODER_EMBEDDING_DIM,
            cfg.MODEL_SELF_HEADS,
            cfg.MODEL_ATT_HEADS,
            emb_dropout=cfg.TRAIN_LSTT_EMB_DROPOUT,
            droppath=cfg.TRAIN_LSTT_DROPPATH,
            lt_dropout=cfg.TRAIN_LSTT_LT_DROPOUT,
            st_dropout=cfg.TRAIN_LSTT_ST_DROPOUT,
            droppath_lst=cfg.TRAIN_LSTT_DROPPATH_LST,
            droppath_scaling=cfg.TRAIN_LSTT_DROPPATH_SCALING,
            intermediate_norm=cfg.MODEL_DECODER_INTERMEDIATE_LSTT,
            return_intermediate=True)

        decoder_indim = cfg.MODEL_ENCODER_EMBEDDING_DIM * \
            (cfg.MODEL_LSTT_NUM * 2 +
             1) if cfg.MODEL_DECODER_INTERMEDIATE_LSTT else cfg.MODEL_ENCODER_EMBEDDING_DIM * 2

        self.decoder = build_decoder(
            decoder,
            in_dim=decoder_indim,
            out_dim=cfg.MODEL_MAX_OBJ_NUM + 1,
            decode_intermediate_input=cfg.MODEL_DECODER_INTERMEDIATE_LSTT,
            hidden_dim=cfg.MODEL_ENCODER_EMBEDDING_DIM,
            shortcut_dims=cfg.MODEL_ENCODER_DIM,
            align_corners=cfg.MODEL_ALIGN_CORNERS)

        self.id_norm = nn.LayerNorm(cfg.MODEL_ENCODER_EMBEDDING_DIM)

        self._init_weight()

    def decode_id_logits(self, lstt_emb, shortcuts):
        n, c, h, w = shortcuts[-1].size()
        decoder_inputs = [shortcuts[-1]]
        for emb in lstt_emb:
            decoder_inputs.append(emb.view(h, w, n, -1).permute(2, 3, 0, 1))
        pred_logit = self.decoder(decoder_inputs, shortcuts)
        return pred_logit

    def get_id_emb(self, x):
        id_emb = self.patch_wise_id_bank(x)
        id_emb = self.id_norm(id_emb.permute(2, 3, 0, 1)).permute(2, 3, 0, 1)
        id_emb = self.id_dropout(id_emb)
        return id_emb



"""
Creates a MobileNetV3 Model as defined in:
Andrew Howard, Mark Sandler, Grace Chu, Liang-Chieh Chen, Bo Chen, Mingxing Tan, Weijun Wang, Yukun Zhu, Ruoming Pang, Vijay Vasudevan, Quoc V. Le, Hartwig Adam. (2019).
Searching for MobileNetV3
arXiv preprint arXiv:1905.02244.
"""

import torch.nn as nn
import math

from castle.aot.utils.learning import freeze_params


def _make_divisible(v, divisor, min_value=None):
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    :param v:
    :param divisor:
    :param min_value:
    :return:
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class h_sigmoid(nn.Module):
    def __init__(self, inplace=True):
        super(h_sigmoid, self).__init__()
        self.relu = nn.ReLU6(inplace=inplace)

    def forward(self, x):
        return self.relu(x + 3) / 6


class h_swish(nn.Module):
    def __init__(self, inplace=True):
        super(h_swish, self).__init__()
        self.sigmoid = h_sigmoid(inplace=inplace)

    def forward(self, x):
        return x * self.sigmoid(x)


class SELayer(nn.Module):
    def __init__(self, channel, reduction=4):
        super(SELayer, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(channel, _make_divisible(channel // reduction, 8)),
            nn.ReLU(inplace=True),
            nn.Linear(_make_divisible(channel // reduction, 8), channel),
            h_sigmoid())

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y


def conv_3x3_bn(inp, oup, stride, norm_layer=nn.BatchNorm2d):
    return nn.Sequential(nn.Conv2d(inp, oup, 3, stride, 1, bias=False),
                         norm_layer(oup), h_swish())


def conv_1x1_bn(inp, oup, norm_layer=nn.BatchNorm2d):
    return nn.Sequential(nn.Conv2d(inp, oup, 1, 1, 0, bias=False),
                         norm_layer(oup), h_swish())


class InvertedResidual(nn.Module):
    def __init__(self,
                 inp,
                 hidden_dim,
                 oup,
                 kernel_size,
                 stride,
                 use_se,
                 use_hs,
                 dilation=1,
                 norm_layer=nn.BatchNorm2d):
        super(InvertedResidual, self).__init__()
        assert stride in [1, 2]

        self.identity = stride == 1 and inp == oup

        if inp == hidden_dim:
            self.conv = nn.Sequential(
                # dw
                nn.Conv2d(hidden_dim,
                          hidden_dim,
                          kernel_size,
                          stride, (kernel_size - 1) // 2 * dilation,
                          dilation=dilation,
                          groups=hidden_dim,
                          bias=False),
                norm_layer(hidden_dim),
                h_swish() if use_hs else nn.ReLU(inplace=True),
                # Squeeze-and-Excite
                SELayer(hidden_dim) if use_se else nn.Identity(),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                norm_layer(oup),
            )
        else:
            self.conv = nn.Sequential(
                # pw
                nn.Conv2d(inp, hidden_dim, 1, 1, 0, bias=False),
                norm_layer(hidden_dim),
                h_swish() if use_hs else nn.ReLU(inplace=True),
                # dw
                nn.Conv2d(hidden_dim,
                          hidden_dim,
                          kernel_size,
                          stride, (kernel_size - 1) // 2 * dilation,
                          dilation=dilation,
                          groups=hidden_dim,
                          bias=False),
                norm_layer(hidden_dim),
                # Squeeze-and-Excite
                SELayer(hidden_dim) if use_se else nn.Identity(),
                h_swish() if use_hs else nn.ReLU(inplace=True),
                # pw-linear
                nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
                norm_layer(oup),
            )

    def forward(self, x):
        if self.identity:
            return x + self.conv(x)
        else:
            return self.conv(x)


class MobileNetV3Large(nn.Module):
    def __init__(self,
                 output_stride=16,
                 norm_layer=nn.BatchNorm2d,
                 width_mult=1.,
                 freeze_at=0):
        super(MobileNetV3Large, self).__init__()
        """
        Constructs a MobileNetV3-Large model
        """
        cfgs = [
            # k, t, c, SE, HS, s
            [3, 1, 16, 0, 0, 1],
            [3, 4, 24, 0, 0, 2],
            [3, 3, 24, 0, 0, 1],
            [5, 3, 40, 1, 0, 2],
            [5, 3, 40, 1, 0, 1],
            [5, 3, 40, 1, 0, 1],
            [3, 6, 80, 0, 1, 2],
            [3, 2.5, 80, 0, 1, 1],
            [3, 2.3, 80, 0, 1, 1],
            [3, 2.3, 80, 0, 1, 1],
            [3, 6, 112, 1, 1, 1],
            [3, 6, 112, 1, 1, 1],
            [5, 6, 160, 1, 1, 2],
            [5, 6, 160, 1, 1, 1],
            [5, 6, 160, 1, 1, 1]
        ]
        self.cfgs = cfgs

        # building first layer
        input_channel = _make_divisible(16 * width_mult, 8)
        layers = [conv_3x3_bn(3, input_channel, 2, norm_layer)]
        # building inverted residual blocks
        block = InvertedResidual
        now_stride = 2
        rate = 1
        for k, t, c, use_se, use_hs, s in self.cfgs:
            if now_stride == output_stride:
                dilation = rate
                rate *= s
                s = 1
            else:
                dilation = 1
                now_stride *= s
            output_channel = _make_divisible(c * width_mult, 8)
            exp_size = _make_divisible(input_channel * t, 8)
            layers.append(
                block(input_channel, exp_size, output_channel, k, s, use_se,
                      use_hs, dilation, norm_layer))
            input_channel = output_channel

        self.features = nn.Sequential(*layers)
        self.conv = conv_1x1_bn(input_channel, exp_size, norm_layer)
        # building last several layers

        self._initialize_weights()

        feature_4x = self.features[0:4]
        feautre_8x = self.features[4:7]
        feature_16x = self.features[7:13]
        feature_32x = self.features[13:]

        self.stages = [feature_4x, feautre_8x, feature_16x, feature_32x]

        self.freeze(freeze_at)

    def forward(self, x):
        xs = []
        for stage in self.stages:
            x = stage(x)
            xs.append(x)
        xs[-1] = self.conv(xs[-1])
        return xs

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                n = m.weight.size(1)
                m.weight.data.normal_(0, 0.01)
                m.bias.data.zero_()

    def freeze(self, freeze_at):
        if freeze_at >= 1:
            for m in self.stages[0][0]:
                freeze_params(m)

        for idx, stage in enumerate(self.stages, start=2):
            if freeze_at >= idx:
                freeze_params(stage)




import cv2
import torch
import numpy as np
from PIL import Image
from .download import download_file
from castle.sam.segment_anything import sam_model_registry, SamAutomaticMaskGenerator
import platform
OS_SYS = platform.uname().system
if OS_SYS == 'Darwin':
    DEFAULT_DEVICE = 'mps'
elif torch.cuda.is_available():
    DEFAULT_DEVICE = 'cuda'
else:
    DEFAULT_DEVICE = 'cpu'


# np.random.seed(200)
# _palette = ((np.random.random((3*255))*0.7+0.3)*255).astype(np.uint8).tolist()
# _palette = [0,0,0]+_palette

# def colorize_mask(pred_mask):
#     save_mask = Image.fromarray(pred_mask.astype(np.uint8))
#     save_mask = save_mask.convert(mode='P')
#     save_mask.putpalette(_palette)
#     save_mask = save_mask.convert(mode='RGB')
#     return np.array(save_mask)

# def draw_points(points, modes, frame):

#     overlay = frame.copy()
#     points = np.array(points)
#     modes = np.array(modes)
#     neg_points = points[np.argwhere(modes == 0)[:, 0]]
#     pos_points = points[np.argwhere(modes == 1)[:, 0]]

#     for i in range(len(neg_points)):
#         point = neg_points[i]
#         cv2.circle(overlay, (point[0], point[1]), 2, (255, 80, 80), -1)

#     for i in range(len(pos_points)):
#         point = pos_points[i]
#         cv2.circle(overlay, (point[0], point[1]), 2, (0, 153, 255), -1)

#     cv2.addWeighted(overlay, 0.5, frame, 0.5, 0, frame)
    
#     return frame

# def merge_frame_and_mask(frame, mask, points = [], modes = [], alpha=0.5):
#     mix = np.array(frame)
#     binary_mask = (mask != 0)
#     foreground = frame * (1-alpha) + colorize_mask(mask) * alpha
#     mix[binary_mask] = foreground[binary_mask]
#     mix = draw_points(points, modes, mix)
#     return mix.astype(np.uint8)


# def mask2img(mask):
#     img = colorize_mask(mask)
#     return img.astype(np.uint8)

class Segmentor:
    def __init__(self, sam_args):
        """
        sam_args:
            sam_checkpoint: path of SAM checkpoint
            generator_args: args for everything_generator
            device: device
        """
        self.device = sam_args["device"]
        self.sam = sam_model_registry[sam_args["model_type"]](checkpoint=sam_args["sam_checkpoint"])
        self.sam.to(device=self.device)
        self.everything_generator = SamAutomaticMaskGenerator(model=self.sam, **sam_args['generator_args'])
        self.interactive_predictor = self.everything_generator.predictor
        self.have_embedded = False
        
    @torch.no_grad()
    def set_image(self, image):
        if not self.have_embedded:
            self.interactive_predictor.set_image(image)
            self.have_embedded = True

    @torch.no_grad()
    def interactive_predict(self, prompts, mode, multimask=True):
        assert self.have_embedded, 'image embedding for sam need be set before predict.'        
        
        if mode == 'point':
            masks, scores, logits = self.interactive_predictor.predict(point_coords=prompts['point_coords'], 
                                point_labels=prompts['point_modes'], 
                                multimask_output=multimask)
        elif mode == 'mask':
            masks, scores, logits = self.interactive_predictor.predict(mask_input=prompts['mask_prompt'], 
                                multimask_output=multimask)
        elif mode == 'point_mask':
            masks, scores, logits = self.interactive_predictor.predict(point_coords=prompts['point_coords'], 
                                point_labels=prompts['point_modes'], 
                                mask_input=prompts['mask_prompt'], 
                                multimask_output=multimask)
                                
        return masks, scores, logits
        
    @torch.no_grad()
    def segment_with_click(self, origin_frame, coords, modes, multimask=True):
        '''
            
            return: 
                mask: one-hot 
        '''
        self.set_image(origin_frame)

        prompts = {
            'point_coords': coords,
            'point_modes': modes,
        }
        masks, scores, logits = self.interactive_predict(prompts, 'point', multimask)
        mask, logit = masks[np.argmax(scores)], logits[np.argmax(scores), :, :]
        prompts = {
            'point_coords': coords,
            'point_modes': modes,
            'mask_prompt': logit[None, :, :]
        }
        masks, scores, logits = self.interactive_predict(prompts, 'point_mask', multimask)
        mask = masks[np.argmax(scores)]

        return mask.astype(np.uint8)
    


class MultiObjectSegmentor():
    def __init__(self, sam_args) -> None:
        self.sam_args = sam_args
        self.click_points = []
        self.click_modes = []
        self.roi_count = 0
        self.next = True
        pass


    def set_frame(self, frame):
        self.frame = frame
        self.pre_mask = np.zeros(frame.shape[:2]).astype(np.uint8)

    def segment_with_click(self, point, mode):
        if self.next:
            self.roi_count += 1
            self.next = False

        self.click_points.append(point)
        self.click_modes.append(mode)

        sam = Segmentor(self.sam_args)
        mask = sam.segment_with_click(
            self.frame, 
            np.array(self.click_points), 
            np.array(self.click_modes)
        )
        del sam
        self.temp_mask = np.array(self.pre_mask)
        self.temp_mask[mask > 0] = self.roi_count
        return self.temp_mask

    def next_roi(self):
        self.next = True
        self.pre_mask = self.temp_mask
        self.click_points = []
        self.click_modes = []

        

    def __del__(self):
        torch.cuda.empty_cache()
   


def download_sa_ckpt(model_type):
    if model_type == 'vit_b':
        ckpt_path = 'ckpt/sam_vit_b_01ec64.pth'
        download_file('https://dl.fbaipublicfiles.com/segment_anything/sam_vit_b_01ec64.pth', ckpt_path)

        return ckpt_path
    else:
        assert False, f"model_type mismatch {model_type}, expect vit_b."


def generate_sa(ckpt_path='', model_type='vit_b', device=''):
    if len(ckpt_path) == 0:
        ckpt_path = download_sa_ckpt(model_type)
    if len(device) == 0:
        device = DEFAULT_DEVICE

    sam_args = {
        'sam_checkpoint': ckpt_path,
        'model_type': model_type,
        'generator_args':{
            'points_per_side': 16,
            'pred_iou_thresh': 0.8,
            'stability_score_thresh': 0.9,
            'crop_n_layers': 1,
            'crop_n_points_downscale_factor': 2,
            'min_mask_region_area': 200,
        },
        'device': device,
    }
    return MultiObjectSegmentor(sam_args)


import os
import shutil


def cp_ckpt(remote_dir="data_wd/youtube_vos_jobs/result", curr_dir="backup"):
    exps = os.listdir(curr_dir)
    for exp in exps:
        print("Exp: ", exp)
        exp_dir = os.path.join(curr_dir, exp)
        stages = os.listdir(exp_dir)
        for stage in stages:
            print("Stage: ", stage)
            stage_dir = os.path.join(exp_dir, stage)
            finals = ["ema_ckpt", "ckpt"]
            for final in finals:
                print("Final: ", final)
                final_dir = os.path.join(stage_dir, final)
                ckpts = os.listdir(final_dir)
                for ckpt in ckpts:
                    if '.pth' not in ckpt:
                        continue
                    curr_ckpt_path = os.path.join(final_dir, ckpt)
                    remote_ckpt_path = os.path.join(remote_dir, exp, stage,
                                                    final, ckpt)
                    if os.path.exists(remote_ckpt_path):
                        os.system('rm {}'.format(remote_ckpt_path))
                    try:
                        shutil.copy(curr_ckpt_path, remote_ckpt_path)
                        print(ckpt, ': OK')
                    except OSError as Inst:
                        print(Inst)
                        print(ckpt, ': Fail')


if __name__ == "__main__":
    cp_ckpt()



import torch
from .resnet import ResNet, Bottleneck

__all__ = ['resnest50', 'resnest101', 'resnest200', 'resnest269']

_url_format = 'https://s3.us-west-1.wasabisys.com/resnest/torch/{}-{}.pth'

_model_sha256 = {
    name: checksum
    for checksum, name in [
        ('528c19ca', 'resnest50'),
        ('22405ba7', 'resnest101'),
        ('75117900', 'resnest200'),
        ('0cc87c48', 'resnest269'),
    ]
}


def short_hash(name):
    if name not in _model_sha256:
        raise ValueError(
            'Pretrained model for {name} is not available.'.format(name=name))
    return _model_sha256[name][:8]


resnest_model_urls = {
    name: _url_format.format(name, short_hash(name))
    for name in _model_sha256.keys()
}


def resnest50(pretrained=False, root='~/.encoding/models', **kwargs):
    model = ResNet(Bottleneck, [3, 4, 6, 3],
                   radix=2,
                   groups=1,
                   bottleneck_width=64,
                   deep_stem=True,
                   stem_width=32,
                   avg_down=True,
                   avd=True,
                   avd_first=False,
                   **kwargs)
    if pretrained:
        model.load_state_dict(
            torch.hub.load_state_dict_from_url(resnest_model_urls['resnest50'],
                                               progress=True,
                                               check_hash=True))
    return model


def resnest101(pretrained=False, root='~/.encoding/models', **kwargs):
    model = ResNet(Bottleneck, [3, 4, 23, 3],
                   radix=2,
                   groups=1,
                   bottleneck_width=64,
                   deep_stem=True,
                   stem_width=64,
                   avg_down=True,
                   avd=True,
                   avd_first=False,
                   **kwargs)
    if pretrained:
        model.load_state_dict(
            torch.hub.load_state_dict_from_url(
                resnest_model_urls['resnest101'],
                progress=True,
                check_hash=True))
    return model


def resnest200(pretrained=False, root='~/.encoding/models', **kwargs):
    model = ResNet(Bottleneck, [3, 24, 36, 3],
                   radix=2,
                   groups=1,
                   bottleneck_width=64,
                   deep_stem=True,
                   stem_width=64,
                   avg_down=True,
                   avd=True,
                   avd_first=False,
                   **kwargs)
    if pretrained:
        model.load_state_dict(
            torch.hub.load_state_dict_from_url(
                resnest_model_urls['resnest200'],
                progress=True,
                check_hash=True))
    return model


def resnest269(pretrained=False, root='~/.encoding/models', **kwargs):
    model = ResNet(Bottleneck, [3, 30, 48, 8],
                   radix=2,
                   groups=1,
                   bottleneck_width=64,
                   deep_stem=True,
                   stem_width=64,
                   avg_down=True,
                   avd=True,
                   avd_first=False,
                   **kwargs)
    if pretrained:
        model.load_state_dict(
            torch.hub.load_state_dict_from_url(
                resnest_model_urls['resnest269'],
                progress=True,
                check_hash=True))
    return model



import os
import json
from re import L
import gradio as gr
import numpy as np
from castle.utils.latent_explorer import Latent
from PIL import Image
import matplotlib.pyplot as plt
from scipy.spatial import KDTree
import io
from castle.utils.video_io import ReadArray
import pandas as pd

umap_config_template = '''[
    {
        "n_neighbors": 30,
        "min_dist": 0.0,
        "n_components": 2
    }
]'''

hdbscan_config_template ='''{
    "n_neighbors": 50,
    "min_dist": 0.0
}'''


dbscan_config_template='''{
    "eps": 2.0
}'''

def padding(mi, mx, scale=1.05):
    mid = (mi + mx) / 2
    d = mx - mi
    return (mid - (d / 2) * scale), (mid + (d / 2) * scale)

class MultiVideos:
    def __init__(self, storage_path, project_name, select_roi_id, bin_size):
        self.source_path = os.path.join(storage_path, project_name, 'sources')
        project_path = os.path.join(storage_path, project_name)
        project_config_path = os.path.join(project_path, 'config.json')
        project_config = json.load(open(project_config_path, 'r'))
        latent_list = [(k,v) for k, v in project_config['latent'].items() if f'ROI_{select_roi_id}' in k]
        latent_dir_path = os.path.join(storage_path, project_name, 'latent')

        self.latents = []
        self.videos_meta = []
        self.bin_size = bin_size


        for it, v in latent_list:
            latent = np.load(os.path.join(latent_dir_path, it))['latent']
            n = (len(latent) // bin_size)* bin_size
            self.latents.append(latent[:n])
            self.videos_meta.append(((len(latent) // bin_size), v))

    
    def bin_index2frame(self, index):

        for vn, v in self.videos_meta:

            if index >= vn:
                index -= vn
                continue
            video_path = os.path.join(self.source_path, v)
            return ReadArray(video_path)[index*self.bin_size + self.bin_size // 2]
        gr.Info('bin_index2frame error')
        return None

    def get_latents(self):
        return Latent(np.array(self.latents), self.bin_size)




class EmbeddingScatterPlot:
    def __init__(self, local_latents):
        data = local_latents.embedding
        self.local_latents = local_latents
        self.data = data
        self.xlim = padding(data[:,0].min(), data[:,0].max())
        self.ylim = padding(data[:,1].min(), data[:,1].max())
        self.selected_point = (np.nan, np.nan)
        self.selected_index = -1
        self.tree = KDTree(data)
    
    def pixel_2_embedding(self, px, py):
        px, py = float(px), float(py)
        ex = (px / self.width) * (self.xlim[1] - self.xlim[0]) + self.xlim[0]
        ey = (py / self.height) * (self.ylim[1] - self.ylim[0]) + self.ylim[0]
        return ex, ey

    def plot(self):
        plt.figure()
        self.local_latents.plot_embedding()
        # plt.scatter(self.data[:,0], self.data[:,1], color='blue')
        plt.scatter(self.selected_point[0], self.selected_point[1], color='red')
        plt.axis('off')
        plt.xlim(self.xlim)
        plt.ylim(self.ylim[1], self.ylim[0])

        buf = io.BytesIO()
        plt.savefig(buf, format='jpeg', bbox_inches='tight', pad_inches=0)
        plt.close()
        buf.seek(0)
        img = Image.open(buf)

        self.width, self.height = img.size

        return img
    
    def click(self, x, y):
        x, y = self.pixel_2_embedding(x, y)
        index = self.near_point(x, y)
        
        self.selected_point = self.data[index]

        self.selected_index = index

        self.selected_index = np.arange(len(self.local_latents.index_mask))[self.local_latents.index_mask][index]
        gr.Info(f'select frame {index}')
        return self.plot()
        
    
    def near_point(self, x, y):
        distance, index = self.tree.query((x, y))
        return index



def embedding_plot_click(mulvideo, Z_plt, evt: gr.SelectData):
    emb_plot = Z_plt.click(evt.index[0], evt.index[1])
    index = Z_plt.selected_index
    # gr.Info(f'index: {index}')
    frame = mulvideo.bin_index2frame(index)
    return emb_plot, frame


def collapse_accordion():
    return gr.update(open=False)

# def init_latent_explorer(storage_path, project_name, select_roi_id, bin_size):
#     if project_name == None:
#         return None
    
#     project_path = os.path.join(storage_path, project_name)
#     project_config_path = os.path.join(project_path, 'config.json')
#     project_config = json.load(open(project_config_path, 'r'))
#     latent_list = [k for k, v in project_config['latent'].items() if f'ROI_{select_roi_id}' in k]
#     latent_dir_path = os.path.join(storage_path, project_name, 'latent')
#     latents = []
#     for it in latent_list:
#         # latent = np.load(f'{latent_dir_path}/{it}')['latent']
#         latent = np.load(os.path.join(latent_dir_path, it))['latent']
#         latents.append(latent)

#     return Latent(np.array(latents), bin_size)


def update_select_cluster_list(latents):
    li = [k for k,v in latents.behavior_name2cluster_id.items()]

    return gr.update(choices=li)



# def select_cluster(latents, cluster_name):
#     return latents.select(selected_cluster=cluster_name)


def generate_embedding(latents, cluster_name, cfg):
    
    try:
        cfg = json.loads(cfg)
    except:
        cfg = dict()
        gr.Info('Json format error')
        return None, None
    local_latents = latents.select(selected_cluster=cluster_name)
    local_latents.build_embedding(cfg)
    Z_plt = EmbeddingScatterPlot(local_latents)
    return local_latents, Z_plt, Z_plt.plot()


def generate_local_cluster(local_latents, method, cfg):
    try:
        cfg = json.loads(cfg)
    except:
        gr.Info('Json format error')
        return None, None
    local_latents.build_cluster(method=method, configs=cfg)
    Z_plt = EmbeddingScatterPlot(local_latents)
    return Z_plt, Z_plt.plot()


def change_cluster_method_template(method):
    if method == 'dbscan':
        return dbscan_config_template
    elif method == 'hdbscan':
        return hdbscan_config_template
    else:
        gr.Info(f'method error got{method}, expect dbscan or hdbscan')



# x.label_cluster(0, 'behavior_bb')
def label_local_cluster(local_latents, cluster_id, cluster_name):
    if len(cluster_name) == 0:
        gr.Info('Name is empty')
        return 
        
    local_latents.label_cluster(cluster_id, cluster_name)  
    gr.Info(f'Name {cluster_id} as {cluster_name}')

def import_info_from_local_latent(storage_path, project_name,latents, local_latents):
    latents.import_local_latent(local_latents)

    fig = plt.figure(figsize=(12, 2))
    latents.plot_syllables()
    plt.tight_layout()

    df1 = {
        'Id': [k for k, v in latents.cluster_meta.items()],
        'Name': [v['name'] for k, v in latents.cluster_meta.items()],
    }
    df1 = pd.DataFrame(df1)
    df2 = {
        'behavior': latents.cluster
    }
    df2 = pd.DataFrame(df2)

    cluster_path = os.path.join(storage_path, project_name, 'cluster')
    os.makedirs(cluster_path, exist_ok=True)


    df1_path = os.path.join(cluster_path, 'id.csv')
    df2_path = os.path.join(cluster_path, 'time_series.csv')
    df1.to_csv(df1_path, index=False)  
    df2.to_csv(df2_path)

    

    return fig, update_select_cluster_list(latents), df1_path, df2_path



def init_mulvideo(storage_path, project_name, select_roi_id, bin_size):
    if project_name == None:
        return None
    
    mulv = MultiVideos(storage_path, project_name, select_roi_id, bin_size)
    return mulv, mulv.get_latents()



def create_cluster_page_ui(storage_path, project_name, cluster_page_tab):
    ui = dict()
    with gr.Accordion('Input setting', visible=False) as ui['cluster_input_accordion']:
        ui['select_roi_id'] = gr.Textbox(label="Enter ROI ID", value="1", info="ex: 1,2,3.", visible=False)
        ui['bin_size'] = gr.Number(label='Time window (frame)', value=1, interactive=True, visible=False)
        ui['reset'] = gr.Button("Initialize", interactive=True, visible=False)
    
    latents = gr.State(None)
    local_latents = gr.State(None)
    local_embedding_plot = gr.State(None)
    mulvideo = gr.State(None)
    with gr.Row(visible=True):
        with gr.Column(scale=5):
            ui['embedding_plot'] = gr.Image(label='Embedding', interactive=False, visible=False)
        with gr.Column(scale=5):
            ui['display'] = gr.Image(label='Display', interactive=False, visible=False)  

    with gr.Row(visible=True):
        with gr.Column(scale=2):
            ui['select_cluster'] = gr.Dropdown(label="Select Cluster",  visible=False,interactive=True)
            ui['umap_config_text'] = gr.Textbox(label='UMAP configs', value=umap_config_template, lines=5, max_lines=30, interactive=True, visible=False)
            ui['umap_run'] = gr.Button("Generate Embedding", interactive=True, visible=False)
        with gr.Column(scale=2):
            ui['cluster_method'] = gr.Dropdown(['dbscan', 'hdbscan'], label='Cluster method',
                              value='dbscan',interactive=True)
            ui['cluster_config_text'] = gr.Textbox(label='Cluster configs', lines=5, max_lines=30, interactive=True, visible=False)
            ui['cluster_run'] = gr.Button("Generate Cluster", interactive=True, visible=False)
        with gr.Column(scale=2):
            ui['label_cluster_id'] = gr.Number(label='Cluster id', interactive=True, visible=False)
            ui['label_cluster_name'] = gr.Textbox(label='Cluster name', interactive=True, visible=False)
            ui['label_cluster_btn'] = gr.Button("Enter", interactive=True, visible=False)
            ui['label_cluster_submit_btn'] = gr.Button("Submit", interactive=True, visible=False)


            
    ui['syllables_plot'] = gr.Plot(label='Syllable', visible=False)
    with gr.Row(visible=True):
        with gr.Column(scale=5):
            ui['behavior_id_csv'] = gr.File(label="Behavior ID", interactive=False, visible=False)
        with gr.Column(scale=5):
            ui['behavior_time_series_csv'] = gr.File(label="Behavior time series", interactive=False, visible=False)

    ui['reset'].click(
        fn=init_mulvideo,
        inputs=[storage_path, project_name, ui['select_roi_id'], ui['bin_size']],
        outputs=[mulvideo, latents]
    )

    ui['select_cluster'].focus(
        fn=update_select_cluster_list,
        inputs=latents,
        outputs=ui['select_cluster']

    )
    # ui['select_cluster'].select(
    #     fn=select_cluster,
    #     inputs=[latents, ui['select_cluster']],
    #     outputs=local_latents,

    # )
    ui['umap_run'].click(
        fn=generate_embedding,
        inputs=[latents, ui['select_cluster'], ui['umap_config_text']],
        outputs=[local_latents, local_embedding_plot, ui['embedding_plot']]
    )

    ui['cluster_method'].select(
        fn=change_cluster_method_template,
        inputs=ui['cluster_method'],
        outputs=ui['cluster_config_text']
    )
    ui['reset'].click(
        fn=change_cluster_method_template,
        inputs=ui['cluster_method'],
        outputs=ui['cluster_config_text']
    )
    ui['reset'].click(
        fn=collapse_accordion,
        outputs=ui['cluster_input_accordion']
    )

    ui['embedding_plot'].select(
        fn=embedding_plot_click,
        inputs=[mulvideo, local_embedding_plot],
        outputs= [ui['embedding_plot'], ui['display']]
    )
    ui['cluster_run'].click(
        fn=generate_local_cluster,
        inputs=[local_latents, ui['cluster_method'], ui['cluster_config_text']],
        outputs=[local_embedding_plot, ui['embedding_plot'] ],
    )
    ui['label_cluster_btn'].click(
        fn=label_local_cluster,
        inputs=[local_latents, ui['label_cluster_id'], ui['label_cluster_name']],
    )
    ui['label_cluster_submit_btn'].click(
        fn=import_info_from_local_latent,
        inputs=[storage_path, project_name, latents, local_latents],
        outputs=[ui['syllables_plot'], ui['select_cluster'], ui['behavior_id_csv'], ui['behavior_time_series_csv']]
    )

    return ui


import math
import torch.nn as nn

from .splat import SplAtConv2d, DropBlock2D
from castle.aot.utils.learning import freeze_params

__all__ = ['ResNet', 'Bottleneck']

_url_format = 'https://s3.us-west-1.wasabisys.com/resnest/torch/{}-{}.pth'

_model_sha256 = {name: checksum for checksum, name in []}


def short_hash(name):
    if name not in _model_sha256:
        raise ValueError(
            'Pretrained model for {name} is not available.'.format(name=name))
    return _model_sha256[name][:8]


resnest_model_urls = {
    name: _url_format.format(name, short_hash(name))
    for name in _model_sha256.keys()
}


class GlobalAvgPool2d(nn.Module):
    def __init__(self):
        """Global average pooling over the input's spatial dimensions"""
        super(GlobalAvgPool2d, self).__init__()

    def forward(self, inputs):
        return nn.functional.adaptive_avg_pool2d(inputs,
                                                 1).view(inputs.size(0), -1)


class Bottleneck(nn.Module):
    """ResNet Bottleneck
    """
    # pylint: disable=unused-argument
    expansion = 4

    def __init__(self,
                 inplanes,
                 planes,
                 stride=1,
                 downsample=None,
                 radix=1,
                 cardinality=1,
                 bottleneck_width=64,
                 avd=False,
                 avd_first=False,
                 dilation=1,
                 is_first=False,
                 rectified_conv=False,
                 rectify_avg=False,
                 norm_layer=None,
                 dropblock_prob=0.0,
                 last_gamma=False):
        super(Bottleneck, self).__init__()
        group_width = int(planes * (bottleneck_width / 64.)) * cardinality
        self.conv1 = nn.Conv2d(inplanes,
                               group_width,
                               kernel_size=1,
                               bias=False)
        self.bn1 = norm_layer(group_width)
        self.dropblock_prob = dropblock_prob
        self.radix = radix
        self.avd = avd and (stride > 1 or is_first)
        self.avd_first = avd_first

        if self.avd:
            self.avd_layer = nn.AvgPool2d(3, stride, padding=1)
            stride = 1

        if dropblock_prob > 0.0:
            self.dropblock1 = DropBlock2D(dropblock_prob, 3)
            if radix == 1:
                self.dropblock2 = DropBlock2D(dropblock_prob, 3)
            self.dropblock3 = DropBlock2D(dropblock_prob, 3)

        if radix >= 1:
            self.conv2 = SplAtConv2d(group_width,
                                     group_width,
                                     kernel_size=3,
                                     stride=stride,
                                     padding=dilation,
                                     dilation=dilation,
                                     groups=cardinality,
                                     bias=False,
                                     radix=radix,
                                     rectify=rectified_conv,
                                     rectify_avg=rectify_avg,
                                     norm_layer=norm_layer,
                                     dropblock_prob=dropblock_prob)
        elif rectified_conv:
            from rfconv import RFConv2d
            self.conv2 = RFConv2d(group_width,
                                  group_width,
                                  kernel_size=3,
                                  stride=stride,
                                  padding=dilation,
                                  dilation=dilation,
                                  groups=cardinality,
                                  bias=False,
                                  average_mode=rectify_avg)
            self.bn2 = norm_layer(group_width)
        else:
            self.conv2 = nn.Conv2d(group_width,
                                   group_width,
                                   kernel_size=3,
                                   stride=stride,
                                   padding=dilation,
                                   dilation=dilation,
                                   groups=cardinality,
                                   bias=False)
            self.bn2 = norm_layer(group_width)

        self.conv3 = nn.Conv2d(group_width,
                               planes * 4,
                               kernel_size=1,
                               bias=False)
        self.bn3 = norm_layer(planes * 4)

        if last_gamma:
            from torch.nn.init import zeros_
            zeros_(self.bn3.weight)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.dilation = dilation
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        if self.dropblock_prob > 0.0:
            out = self.dropblock1(out)
        out = self.relu(out)

        if self.avd and self.avd_first:
            out = self.avd_layer(out)

        out = self.conv2(out)
        if self.radix == 0:
            out = self.bn2(out)
            if self.dropblock_prob > 0.0:
                out = self.dropblock2(out)
            out = self.relu(out)

        if self.avd and not self.avd_first:
            out = self.avd_layer(out)

        out = self.conv3(out)
        out = self.bn3(out)
        if self.dropblock_prob > 0.0:
            out = self.dropblock3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):
    """ResNet Variants
    Parameters
    ----------
    block : Block
        Class for the residual block. Options are BasicBlockV1, BottleneckV1.
    layers : list of int
        Numbers of layers in each block
    classes : int, default 1000
        Number of classification classes.
    dilated : bool, default False
        Applying dilation strategy to pretrained ResNet yielding a stride-8 model,
        typically used in Semantic Segmentation.
    norm_layer : object
        Normalization layer used in backbone network (default: :class:`mxnet.gluon.nn.BatchNorm`;
        for Synchronized Cross-GPU BachNormalization).
    Reference:
        - He, Kaiming, et al. "Deep residual learning for image recognition." Proceedings of the IEEE conference on computer vision and pattern recognition. 2016.
        - Yu, Fisher, and Vladlen Koltun. "Multi-scale context aggregation by dilated convolutions."
    """

    # pylint: disable=unused-variable
    def __init__(self,
                 block,
                 layers,
                 radix=1,
                 groups=1,
                 bottleneck_width=64,
                 num_classes=1000,
                 dilated=False,
                 dilation=1,
                 deep_stem=False,
                 stem_width=64,
                 avg_down=False,
                 rectified_conv=False,
                 rectify_avg=False,
                 avd=False,
                 avd_first=False,
                 final_drop=0.0,
                 dropblock_prob=0,
                 last_gamma=False,
                 norm_layer=nn.BatchNorm2d,
                 freeze_at=0):
        self.cardinality = groups
        self.bottleneck_width = bottleneck_width
        # ResNet-D params
        self.inplanes = stem_width * 2 if deep_stem else 64
        self.avg_down = avg_down
        self.last_gamma = last_gamma
        # ResNeSt params
        self.radix = radix
        self.avd = avd
        self.avd_first = avd_first

        super(ResNet, self).__init__()
        self.rectified_conv = rectified_conv
        self.rectify_avg = rectify_avg
        if rectified_conv:
            from rfconv import RFConv2d
            conv_layer = RFConv2d
        else:
            conv_layer = nn.Conv2d
        conv_kwargs = {'average_mode': rectify_avg} if rectified_conv else {}
        if deep_stem:
            self.conv1 = nn.Sequential(
                conv_layer(3,
                           stem_width,
                           kernel_size=3,
                           stride=2,
                           padding=1,
                           bias=False,
                           **conv_kwargs),
                norm_layer(stem_width),
                nn.ReLU(inplace=True),
                conv_layer(stem_width,
                           stem_width,
                           kernel_size=3,
                           stride=1,
                           padding=1,
                           bias=False,
                           **conv_kwargs),
                norm_layer(stem_width),
                nn.ReLU(inplace=True),
                conv_layer(stem_width,
                           stem_width * 2,
                           kernel_size=3,
                           stride=1,
                           padding=1,
                           bias=False,
                           **conv_kwargs),
            )
        else:
            self.conv1 = conv_layer(3,
                                    64,
                                    kernel_size=7,
                                    stride=2,
                                    padding=3,
                                    bias=False,
                                    **conv_kwargs)
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block,
                                       64,
                                       layers[0],
                                       norm_layer=norm_layer,
                                       is_first=False)
        self.layer2 = self._make_layer(block,
                                       128,
                                       layers[1],
                                       stride=2,
                                       norm_layer=norm_layer)
        if dilated or dilation == 4:
            self.layer3 = self._make_layer(block,
                                           256,
                                           layers[2],
                                           stride=1,
                                           dilation=2,
                                           norm_layer=norm_layer,
                                           dropblock_prob=dropblock_prob)
        elif dilation == 2:
            self.layer3 = self._make_layer(block,
                                           256,
                                           layers[2],
                                           stride=2,
                                           dilation=1,
                                           norm_layer=norm_layer,
                                           dropblock_prob=dropblock_prob)
        else:
            self.layer3 = self._make_layer(block,
                                           256,
                                           layers[2],
                                           stride=2,
                                           norm_layer=norm_layer,
                                           dropblock_prob=dropblock_prob)

        self.stem = [self.conv1, self.bn1]
        self.stages = [self.layer1, self.layer2, self.layer3]

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, norm_layer):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

        self.freeze(freeze_at)

    def _make_layer(self,
                    block,
                    planes,
                    blocks,
                    stride=1,
                    dilation=1,
                    norm_layer=None,
                    dropblock_prob=0.0,
                    is_first=True):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            down_layers = []
            if self.avg_down:
                if dilation == 1:
                    down_layers.append(
                        nn.AvgPool2d(kernel_size=stride,
                                     stride=stride,
                                     ceil_mode=True,
                                     count_include_pad=False))
                else:
                    down_layers.append(
                        nn.AvgPool2d(kernel_size=1,
                                     stride=1,
                                     ceil_mode=True,
                                     count_include_pad=False))
                down_layers.append(
                    nn.Conv2d(self.inplanes,
                              planes * block.expansion,
                              kernel_size=1,
                              stride=1,
                              bias=False))
            else:
                down_layers.append(
                    nn.Conv2d(self.inplanes,
                              planes * block.expansion,
                              kernel_size=1,
                              stride=stride,
                              bias=False))
            down_layers.append(norm_layer(planes * block.expansion))
            downsample = nn.Sequential(*down_layers)

        layers = []
        if dilation == 1 or dilation == 2:
            layers.append(
                block(self.inplanes,
                      planes,
                      stride,
                      downsample=downsample,
                      radix=self.radix,
                      cardinality=self.cardinality,
                      bottleneck_width=self.bottleneck_width,
                      avd=self.avd,
                      avd_first=self.avd_first,
                      dilation=1,
                      is_first=is_first,
                      rectified_conv=self.rectified_conv,
                      rectify_avg=self.rectify_avg,
                      norm_layer=norm_layer,
                      dropblock_prob=dropblock_prob,
                      last_gamma=self.last_gamma))
        elif dilation == 4:
            layers.append(
                block(self.inplanes,
                      planes,
                      stride,
                      downsample=downsample,
                      radix=self.radix,
                      cardinality=self.cardinality,
                      bottleneck_width=self.bottleneck_width,
                      avd=self.avd,
                      avd_first=self.avd_first,
                      dilation=2,
                      is_first=is_first,
                      rectified_conv=self.rectified_conv,
                      rectify_avg=self.rectify_avg,
                      norm_layer=norm_layer,
                      dropblock_prob=dropblock_prob,
                      last_gamma=self.last_gamma))
        else:
            raise RuntimeError("=> unknown dilation size: {}".format(dilation))

        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(
                block(self.inplanes,
                      planes,
                      radix=self.radix,
                      cardinality=self.cardinality,
                      bottleneck_width=self.bottleneck_width,
                      avd=self.avd,
                      avd_first=self.avd_first,
                      dilation=dilation,
                      rectified_conv=self.rectified_conv,
                      rectify_avg=self.rectify_avg,
                      norm_layer=norm_layer,
                      dropblock_prob=dropblock_prob,
                      last_gamma=self.last_gamma))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        xs = []

        x = self.layer1(x)
        xs.append(x)  # 4X
        x = self.layer2(x)
        xs.append(x)  # 8X
        x = self.layer3(x)
        xs.append(x)  # 16X
        # Following STMVOS, we drop stage 5.
        xs.append(x)  # 16X

        return xs

    def freeze(self, freeze_at):
        if freeze_at >= 1:
            for m in self.stem:
                freeze_params(m)

        for idx, stage in enumerate(self.stages, start=2):
            if freeze_at >= idx:
                freeze_params(stage)
